//
//  dds.c
//  
//
//  Created by Matt Stern on 10/29/13.
//
//

#include <util/delay.h>
#include <math.h>

#include "midi_lib.c"

#define PORTDIR DDRC
#define PORTDATA PORTC
#define PORTIN PINC

//Important:
//Assumes that Timer0 is operating at full clock rate
//and calls midiISR() every 1 msec

//The buttons pushed: bit i set iff button i pressed
uint16_t buttons;

// for a 16-bit DDS accumulator, running at 8000 Hz:
// increment = 2^16*f_out/8e3
const float INCREMENT_CONSTANT = 8.192;

// DDS variables
//DDS variables
#define MAX_CH 5
#define MAX_NOTES 3
#define max_amp 32767

//THE LAST CHANNEL IS RESERVED FOR BUTTON PRESSES ONLY
volatile uint16_t acc_main[MAX_CH][MAX_NOTES], acc_fm1[MAX_CH][MAX_NOTES];
volatile uint8_t  high_main[MAX_CH][MAX_NOTES], high_fm1[MAX_CH][MAX_NOTES], decay_fm1[MAX_CH][MAX_NOTES], decay_main[MAX_CH][MAX_NOTES], depth_fm1[MAX_CH][MAX_NOTES], rise_main[MAX_CH][MAX_NOTES];
volatile uint8_t  sel_main[MAX_CH][MAX_NOTES], sel_fm1[MAX_CH][MAX_NOTES];
volatile uint16_t inc_main[MAX_CH][MAX_NOTES], inc_fm1[MAX_CH][MAX_NOTES], amp_main[MAX_CH][MAX_NOTES], amp_fm1[MAX_CH][MAX_NOTES];
volatile uint16_t rise_phase_main[MAX_CH][MAX_NOTES], amp_rise_main[MAX_CH][MAX_NOTES], amp_fall_main[MAX_CH][MAX_NOTES];
volatile uint8_t  pluck[MAX_CH][MAX_NOTES];
volatile uint8_t  fm1[MAX_CH][MAX_NOTES];
volatile uint8_t  channel_on[MAX_CH][MAX_NOTES];

#define CLOCK_FOR_FM 255;
volatile uint8_t clock_ticks_fm;

//Debouncing interval
#define DEBOUNCE_PD_MSEC 4
volatile uint8_t debounce_msec = DEBOUNCE_PD_MSEC;
volatile uint8_t debounce_flag = 0;

//FM parameters table and constants
float fm_parameters[16][5];
#define SEL_MAIN 0
#define SEL_FM 1
#define INC_FM 2
#define DECAY_FM 3
#define DEPTH_FM 4

//The current channel and instrument being recorded from button board
uint8_t cur_chan = 0;
uint8_t cur_instrument = 0;

//Wave table for DDS and constants for different waveforms
int8_t waveTable[3][256];
#define SINE 0
#define SAW 1
#define TRI 2

uint16_t midiIncrTable[108]; //Maps each MIDI note to the proper DDS increment
uint16_t buttonNotes[16];  //Maps each button to a MIDI note

uint16_t notes_playing = 0; //To limit number of notes recorded

int32_t find_note_off(uint16_t ch, uint16_t inc){
	//Finds the current note within a channel that is playing `increment`
	for (uint16_t note = 0; note < MAX_NOTES; note++){
		if (inc_main[ch][note] == inc && channel_on[ch][note]){
			return note;
		}
	}
	return -1;
}

int32_t find_free_note(uint16_t ch, uint16_t inc){
	//Finds a free note to use in channel `ch`. Returns -1 if all 3
	//notes are in use or the note is already playing
	if (find_note_off(ch, inc) != -1){
		return -1;
	}
	for (uint16_t note = 0; note < MAX_NOTES; note++){
		if (!channel_on[ch][note]) {
			return note;
		}
	}
	return -1;
}

void changeButton(uint8_t button_num, uint8_t button_val){
    //Changes button `button_num` to play MIDI note `button_val`
	buttonNotes[button_num] = button_val;
}

void changeInstrument(uint8_t channel, uint8_t instr){
    //Change the FM synthesis parameters of MIDI channel `channel
    //to play MIDI instrument number `instr`
	for (uint8_t i = 0; i < MAX_NOTES; i++){
		sel_main[channel][i] = fm_parameters[instr][SEL_MAIN];
		inc_fm1[channel][i] = fm_parameters[instr][INC_FM];
		depth_fm1[channel][i] = fm_parameters[instr][DEPTH_FM];
		decay_fm1[channel][i] = fm_parameters[instr][DECAY_FM];
		sel_fm1[channel][i] = fm_parameters[instr][SEL_FM];
	}
}

void initParameters(void) {
    //Init table of FM synthesis parameters for each MIDI instrument
    //that we have created
	fm_parameters[0][SEL_MAIN] = SINE;
	fm_parameters[0][SEL_FM] = SINE;
	fm_parameters[0][INC_FM] = 350.0 * INCREMENT_CONSTANT;
	fm_parameters[0][DECAY_FM] = 5;
	fm_parameters[0][DEPTH_FM] = 8;

	fm_parameters[1][SEL_MAIN] = SINE;
	fm_parameters[1][SEL_FM] = SINE;
	fm_parameters[1][INC_FM] = 500.0 * INCREMENT_CONSTANT;
	fm_parameters[1][DECAY_FM] = 14;
	fm_parameters[1][DEPTH_FM] = 10;

	fm_parameters[2][SEL_MAIN] = SINE;
	fm_parameters[2][SEL_FM] = SINE;
	fm_parameters[2][INC_FM] = 1000.0 * INCREMENT_CONSTANT;
	fm_parameters[2][DECAY_FM] = 11;
	fm_parameters[2][DEPTH_FM] = 10;

	fm_parameters[3][SEL_MAIN] = SINE;
	fm_parameters[3][SEL_FM] = TRI;
	fm_parameters[3][INC_FM] = 150.0 * INCREMENT_CONSTANT;
	fm_parameters[3][DECAY_FM] = 6;
	fm_parameters[3][DEPTH_FM] = 8;

	fm_parameters[4][SEL_MAIN] = SINE;
	fm_parameters[4][SEL_FM] = TRI;
	fm_parameters[4][INC_FM] = 1000.0 * INCREMENT_CONSTANT;
	fm_parameters[4][DECAY_FM] = 6;
	fm_parameters[4][DEPTH_FM] = 8;

	fm_parameters[5][SEL_MAIN] = SINE;
	fm_parameters[5][SEL_FM] = SINE;
	fm_parameters[5][INC_FM] = 261.0 * INCREMENT_CONSTANT;
	fm_parameters[5][DECAY_FM] = 6;
	fm_parameters[5][DEPTH_FM] = 7;

	fm_parameters[6][SEL_MAIN] = SINE;
	fm_parameters[6][SEL_FM] = SINE;
	fm_parameters[6][INC_FM] = 750.0 * INCREMENT_CONSTANT;
	fm_parameters[6][DECAY_FM] = 3;
	fm_parameters[6][DEPTH_FM] = 8;

	fm_parameters[7][SEL_MAIN] = SINE;
	fm_parameters[7][SEL_FM] = SINE;
	fm_parameters[7][INC_FM] = 300.0 * INCREMENT_CONSTANT;
	fm_parameters[7][DECAY_FM] = 6;
	fm_parameters[7][DEPTH_FM] = 8;

	fm_parameters[8][SEL_MAIN] = SINE;
	fm_parameters[8][SEL_FM] = SINE;
	fm_parameters[8][INC_FM] = 50.0 * INCREMENT_CONSTANT;
	fm_parameters[8][DECAY_FM] = 10;
	fm_parameters[8][DEPTH_FM] = 5;

	fm_parameters[9][SEL_MAIN] = SINE;
	fm_parameters[9][SEL_FM] = SINE;
	fm_parameters[9][INC_FM] = 600.0 * INCREMENT_CONSTANT;
	fm_parameters[9][DECAY_FM] = 6;
	fm_parameters[9][DEPTH_FM] = 8;

	fm_parameters[10][SEL_MAIN] = SINE;
	fm_parameters[10][SEL_FM] = SINE;
	fm_parameters[10][INC_FM] = 1000.0 * INCREMENT_CONSTANT;
	fm_parameters[10][DECAY_FM] = 6;
	fm_parameters[10][DEPTH_FM] = 8;

	fm_parameters[11][SEL_MAIN] = SINE;
	fm_parameters[11][SEL_FM] = SINE;
	fm_parameters[11][INC_FM] = 850.0 * INCREMENT_CONSTANT;
	fm_parameters[11][DECAY_FM] = 4;
	fm_parameters[11][DEPTH_FM] = 6;

	fm_parameters[12][SEL_MAIN] = SINE;
	fm_parameters[12][SEL_FM] = SINE;
	fm_parameters[12][INC_FM] = 350.0 * INCREMENT_CONSTANT;
	fm_parameters[12][DECAY_FM] = 5;
	fm_parameters[12][DEPTH_FM] = 8;

	fm_parameters[13][SEL_MAIN] = SAW;
	fm_parameters[13][SEL_FM] = SINE;
	fm_parameters[13][INC_FM] = 350.0 * INCREMENT_CONSTANT;
	fm_parameters[13][DECAY_FM] = 5;
	fm_parameters[13][DEPTH_FM] = 8;

	fm_parameters[14][SEL_MAIN] = SAW;
	fm_parameters[14][SEL_FM] = SAW;
	fm_parameters[14][INC_FM] = 800.0 * INCREMENT_CONSTANT;
	fm_parameters[14][DECAY_FM] = 5;
	fm_parameters[14][DEPTH_FM] = 8;

	fm_parameters[15][SEL_MAIN] = SINE;
	fm_parameters[15][SEL_FM] = SAW;
	fm_parameters[15][INC_FM] = 300.0 * INCREMENT_CONSTANT;
	fm_parameters[15][DECAY_FM] = 4;
	fm_parameters[15][DEPTH_FM] = 7;
}

void initTables(void){
    //Initiaze all tables required for DDS
	for (uint16_t i = 0; i < MAX_CH; i++){
		for (uint16_t j = 0; j < MAX_NOTES; j++){
			channel_on[i][j]  = 0;
			inc_main[i][j] = 0;
		}
	}

    //Initialize wave table
   	for (uint16_t i = 0; i < 256; i++){
   		waveTable[0][i] = (int8_t)(127.0 * sin(6.283*((float)i)/256.0)) ;
		waveTable[1][i] = (int8_t)(( (i<17)? i*7 : (i<241)? 128-i : i*7-1792 )) ;
		waveTable[2][i] = (int8_t) (( (i<64)? i : (i<192)? 128-i : i-256 )) ;
   	} 

    //The lowest octave of each MIDI note. Note that the frequency of
    //Xn = 2 * X(n-1), where X is a note name and n is an octave number
	float baseNotes[12] = {
		27.500, //A0
		29.135, //A#0
		30.868, //B0
		32.703, //C1
		34.648, //C#1
		36.708, //D1
		38.891, //D#1
		41.203, //E1
		43.654, //F1
		46.249, //F#1
		48.999, //G1
		51.913  //G#1

	};

    //Initialize MIDI increment table by multipying the base notes
    //by a constant to derive each note
	uint8_t remainder;
	uint8_t multiplier = 1;	
	for (uint16_t i=0; i < 88; i++){
		remainder = i % 12;
		if (i && !remainder){
			multiplier <<= 1;
		}
		midiIncrTable[i + 21] = (uint16_t) ( baseNotes[i % 12] * ((float) multiplier) * INCREMENT_CONSTANT);
	}
}

//FSM states for debouncing keypad
#define NO_PUSH  		0
#define DEBOUNCE  		1
#define STILL_PRESSED	2
#define DEBOUNCE_REL	3

//FSM variables
uint8_t  curState;
uint16_t maybe;

void keypad(void){
    //Reads in the values of PORTDIR and stores them in `buttons`
	PORTDIR  = 0x00;
	PORTDATA = 0xff;
	buttons = 0xff & ~(PORTIN); //since buttons is 16-bit for the future
}

void playButtonPress(void){
    //Turns notes on and off according to the current state of `buttons`
    
	//Update DDS playback with current buttons state;
	uint16_t mask = 1;

	//No button was pressed, turn all notes off
	if (!buttons){
		for (uint16_t i = 0; i < MAX_NOTES; i++){
			channel_on[MAX_CH-1][i] = 0;
		}
	//Turn on notes corresponding to button presses capped
	//at MAX_NOTES notes
	} else {
		for (uint8_t i = 0; i < 16; i++){
			if (buttons & mask){
				uint16_t inc = midiIncrTable[buttonNotes[i]];
				int32_t free_note= find_free_note(MAX_CH-1, inc);
				if (free_note != -1){
					inc_main[MAX_CH-1][free_note]  = inc;
					channel_on[MAX_CH-1][free_note] = 1;
					pluck[MAX_CH-1][free_note] = 1;
				}
			}
			mask <<= 1;
		}
	}
}


void _recordButtonPress(uint8_t event, uint8_t channel){
	//Records a MIDI event `event` on channel `channel` to the
    //curent MIDI stream that is recording
	uint16_t mask = 1;
	for (uint8_t i = 0; i < 16; i++){
		if (maybe & mask){
			midiEvent(event, channel, buttonNotes[i], 0x00);
		}
		mask <<= 1;
	}
}

void noteOn(uint8_t channel){
	//Records a note on MIDI event channel `channel` to the
    //curent MIDI stream that is recording
	if (notes_playing < MAX_NOTES-1){
		notes_playing++;
		_recordButtonPress(0x9, channel);
	}
}

void noteOff(uint8_t channel){
	//Records a note off MIDI event channel `channel` to the
    //curent MIDI stream that is recording
	if (notes_playing){
		notes_playing--;
		_recordButtonPress(0x8, channel);
	}
}

void updateKeypad(void){
	//Debouncing FSM for keypad. Plays back key presses and records
	//them in the current MIDI stream
    while(!debounce_flag);
	debounce_flag = 0;
	keypad();
    switch (curState){
        case NO_PUSH:
            if (buttons != 0){
                maybe = buttons;
                curState = DEBOUNCE;
            }
            break;
        case DEBOUNCE:
            if (buttons == maybe){
                curState = STILL_PRESSED;
				noteOn(cur_chan);
				playButtonPress();
            } else {
                curState = NO_PUSH;
            }
            break;
        case STILL_PRESSED:
            if (buttons != maybe){
                curState = DEBOUNCE_REL;
            }
            break;
        case DEBOUNCE_REL:
            if (buttons == maybe){
                curState = STILL_PRESSED;
            } else {
				noteOff(cur_chan);
				playButtonPress();
                curState = NO_PUSH;
            }
            break;
        default:
            break;
    }
}

int32_t processQFM(volatile uint8_t* stream, int32_t idx, int32_t num){
    //Executes all MIDI events within one quarter frame
    //Precondition: idx_ptr is currently at the MSB of a QFM
    //Returns the index of the next QFM
    
    //Advance idx to the first event
    idx += 2;
    //Find the idx of the next QFM
    int32_t next_QFM_idx = indexOfQFM(stream, idx, num);
    
    //There are no more QFMs in the stream
    if (next_QFM_idx != -1){
        //Process all events in between
        while (idx < next_QFM_idx){
            uint8_t event   = stream[idx] & 0xf0;
            uint8_t channel = stream[idx] & 0x0f;
			uint8_t param_1 = stream[idx+1];
			int32_t off_note, free_note;
			uint16_t inc;

			//Reserve last channel for button playback
			if (channel >= MAX_CH-1){
				idx += 3;
				continue;
			}
        
            switch (event){
                case 0x80: //Note off
					inc = midiIncrTable[param_1];
					off_note = find_note_off(channel, inc);
					if (off_note != -1){
						channel_on[channel][off_note] = 0;
					}
                    break;
                case 0x90: //Note on
					inc = midiIncrTable[param_1];
					free_note = find_free_note(channel, inc);
					if (free_note != -1){
						inc_main[channel][free_note] = inc;
						channel_on[channel][free_note] = 1;
						pluck[channel][free_note] = 1;
					}
					break;
				case 0xc0: //Instrument
					changeInstrument(channel, param_1);
                    break;
                default:
                    break;
            }
            
            idx += 3;
        }
    }
    
    return next_QFM_idx;
}

void initDDS(void){
    // turn on fast PWM and OC0A output
    // at full clock rate, toggle OC0A (pin B3)
    // 16 microsec per PWM cycle sample time
    TCCR0A = (1<<COM0A0) | (1<<COM0A1) | (1<<WGM00) | (1<<WGM01);
}

