//
//  engine.c
//  
//
//  Created by Matt Stern on 10/29/13.
//
// References: 
// http://people.ece.cornell.edu/land/courses/ece4760/Math/GCC644/FM_synth/FM_synth_2.c
//

#define F_CPU 16000000
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <math.h>
#include <inttypes.h>
#include <stdio.h>

#include "uart.h"
#include "uart.c"

#include "dds.c"
#include "ir_comm.c"
#include "teacher.c"

//UART file descriptor
//uart_putchar and uart_getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

//Number of Timer1 overflows per msec
#define CLOCK_PD 8

//Time variables for 1 msec timebase
volatile uint8_t clock_ticks;

//User mode variables
volatile uint8_t is_recording  = 1;
volatile uint8_t is_playback   = 0;
volatile uint8_t is_learning   = 0;
volatile uint8_t play_next_QFM = 0;
volatile uint8_t finished_recording = 0;
volatile uint8_t finished_playback = 0;
volatile uint8_t ir_comm_mode = 0; // SET TO 1 IF WE WANT TO COMMUNICATE IR.

//Number of channels in playback. Set by setNumChannels
uint8_t _num_channels_shamt = 4; //Must be a power of two to make ISR deadline

void setSampleNotes(void){
    //Defaults the button mapping to a C scale
	buttonNotes[0x0] = 60; //C4
	buttonNotes[0x1] = 62; //D4
	buttonNotes[0x2] = 64; //E4
	buttonNotes[0x3] = 65; //F4
	buttonNotes[0x4] = 67; //G4
	buttonNotes[0x5] = 69; //A4
	buttonNotes[0x6] = 71; //B4
	buttonNotes[0x7] = 72; //C5
	buttonNotes[0x8] = 74; //D5
	buttonNotes[0x9] = 76; //E5
	buttonNotes[0xa] = 77; //F5
	buttonNotes[0xb] = 79; //G5
	buttonNotes[0xc] = 81; //A5
	buttonNotes[0xd] = 83; //B5
	buttonNotes[0xe] = 84; //C6
	buttonNotes[0xf] = 86; //D6
}

void resetAll(){
    //Resets all state for MIDI recording
    
	//Remove the header chunk to invalidate the stream
	for(uint8_t i = 0; i < MIDI_STREAM_OVHD; i++){
		pb_stream[i] = 0;
		midi_stream[i] = 0;
	}

	//Reset the idxs
	pb_stream_idx = 0;
	midi_stream_idx = 0;

	//Reset the channel
	num_chan = 0;
	cur_instrument = 0;
}

void saveRecording(){
	//Copies midi_stream to pb_stream
	uint32_t len = getTrackLen(midi_stream);
	memcpy((void *) pb_stream, (void *) midi_stream, len + MIDI_STREAM_OVHD);
	pb_stream_len = MIDI_STREAM_OVHD + getTrackLen(pb_stream);
}

//------------------------------------------------------
// Code for generating songs from code
//------------------------------------------------------

void playNote(uint8_t channel, uint8_t note, uint16_t on_dur, uint16_t off_dur){
	//Adds MIDI note `note` to channel `channel` with duration `on_dur
    //msec, followed by `off_dur` msec of dead space in `midi_stream`
    
	midiEvent(0x9, channel, note, 0xbb);
	for (uint16_t j = 0; j < on_dur; j++){
       incrementSMPTE();
        if (isQuarterFrame()) addQuarterFrameMsg();
    }
	midiEvent(0x8, channel, note, 0xbb);
	for (uint16_t j = 0; j < off_dur; j++){
       incrementSMPTE();
        if (isQuarterFrame()) addQuarterFrameMsg();
    }
}

void playRest(uint8_t channel, uint16_t dur){
    //Adds `dur` msec of dead space to `midi_stream`
	for (uint16_t j = 0; j < dur; j++){
       incrementSMPTE();
        if (isQuarterFrame()) addQuarterFrameMsg();
    }
}

void quarterRest(uint8_t channel){
    //Adds a 120 BPM quarter rest to `midi_stream`
	playRest(channel, 500);
}

void eighthRest(uint8_t channel){
    //Adds a 120 BPM eighth rest to `midi_stream`
	playRest(channel, 250);
}
	
void eighthNote(uint8_t channel, uint8_t note){
    //Adds a 120 BPM eighth note to `midi_stream`
	playNote(channel, note, 220, 30);
}

void quarterNote(uint8_t channel, uint8_t note){
    //Adds a 120 BPM quarter note to `midi_stream`
	playNote(channel, note, 440, 60);
}

void quarterNoteTriplet(uint8_t channel, uint8_t note){
    //Adds 1 of 3 120 BPM quarter note triplets to `midi_stream`
	playNote(channel, note, 300, 33);
}

//------------------------------------------------------
// End code for generating songs from code
//------------------------------------------------------

void genSampleMidi(void){
    //Generates the code that plays "Twinkle Twinkle Little Star"
    
	resetAll();
	setSampleNotes();

    //uint8_t toggle = 1;
    addQuarterFrameMsg();
	midiEvent(0xc, 0x0, 0x0, 0x0);

	eighthNote(0, 60);
	eighthNote(0, 60);
	eighthNote(0, 67);
	eighthNote(0, 67);
	eighthNote(0, 69);
	eighthNote(0, 69);
	quarterNote(0, 67);

	eighthNote(0, 65);
	eighthNote(0, 65);
	eighthNote(0, 64);
	eighthNote(0, 64);
	eighthNote(0, 62);
	eighthNote(0, 62);
	quarterNote(0, 60);

	eighthNote(0, 67);
	eighthNote(0, 67);
	eighthNote(0, 65);
	eighthNote(0, 65);
	eighthNote(0, 64);
	eighthNote(0, 64);
	quarterNote(0, 62);

	eighthNote(0, 67);
	eighthNote(0, 67);
	eighthNote(0, 65);
	eighthNote(0, 65);
	eighthNote(0, 64);
	eighthNote(0, 64);
	quarterNote(0, 62);

    finalizeMidiStream();
	saveRecording();
}


void genSampleMidi2(void){
    //Generates the code that plays "Rugrats Theme"
	resetAll();
	setSampleNotes();

    addQuarterFrameMsg();
	midiEvent(0xc, 0x0, 0x0, 0x0);

   	eighthNote(0, 60);
	eighthNote(0, 62);
	eighthNote(0, 64);
	quarterNoteTriplet(0, 65);
	quarterNoteTriplet(0, 67);
	quarterNote(0, 69); //450 133
	quarterNote(0, 67);
	eighthNote(0, 64);
	quarterNoteTriplet(0, 65);
	quarterNoteTriplet(0, 64);
	quarterNoteTriplet(0, 62);

	eighthRest(0);

	eighthNote(0, 60);
	eighthNote(0, 62);
	eighthNote(0, 64);
	quarterNoteTriplet(0, 65);
	quarterNoteTriplet(0, 71);
	quarterNote(0, 72);
	quarterNote(0, 71);
	eighthNote(0, 67);
	quarterNoteTriplet(0, 69);
	quarterNoteTriplet(0, 67);
	quarterNoteTriplet(0, 65);
    
    finalizeMidiStream();
	saveRecording();
}

void beginRecording(){
    //Set state to begin recording in main loop
	midi_stream_idx = 0;
	addQuarterFrameMsg();
	midiEvent(0xc, cur_chan, cur_instrument, 0x0);
	clearSMPTE();
	is_recording = 1;
}

void beginPlayback(){
    //Set state to begin playback in main loop
    pb_stream_len = getTrackLen(pb_stream) + MIDI_STREAM_OVHD;
    pb_stream_idx = MIDI_STREAM_OVHD;
	clearSMPTE();
	is_playback = 1;
	finished_playback = 0;
}

void serialComm(){
    // Gets user input from UART and updates state variables accordingly
	uint16_t param_1, param_2;
    char cmd[7];

	cli(); //Disble interrupts while in UART mode
    while(1){
        fprintf(stdout, "Enter a command: 'inst', 'but', 'rec', 'play', 'learn', 'erase', 'send', or 'receive'.\n>");
		fscanf(stdin, "%s", cmd);
        if (strcmp(cmd, "inst") == 0){
			fprintf(stdout, "Enter an instrument number 0-15 for this channel\n>");
			fscanf(stdin, "%i", &param_1) ;
			if (param_1 < 16 && param_2 >=0){
				cur_instrument = (uint8_t) param_1;
				changeInstrument(MAX_CH-1, param_1);
			} else {
				fprintf(stdout, "Invalid input\n");
			}
		} else if (strcmp(cmd, "but") == 0){
				fprintf(stdout, "Map buttons 0-15 to MIDI notes 21 - 108.\nEnter command of form: <butnum> <note>.\n>");
				fscanf(stdin, "%i %i", &param_1, &param_2) ;
				if (param_1 < 16 && param_2 >= 21 && param_2 <= 108){
					changeButton((uint8_t) param_1, (uint8_t) param_2);
				} else {
					fprintf(stdout, "Invalid input\n");
				}
		} else if (strcmp(cmd, "rec") == 0){
			is_recording = 1;
			if (validHeaderChunk(pb_stream)){
				is_playback = 1;
			}
			break;
		} else if (strcmp(cmd, "play") == 0){
			is_playback  = 1;
			is_recording = 0;
			break;
		} else if (strcmp(cmd, "learn") == 0){
			//Put a sample song to `pb_stream`
			printf("Select a track to learn.\n");
			printf("Choices are 'twinkle' and 'rugrats'\n");
			printf(">");
			fscanf(stdin, "%s", cmd);
			if (strcmp(cmd, "twinkle") == 0){
				genSampleMidi();
			} else if (strcmp(cmd, "rugrats") == 0){
				genSampleMidi2();
			} else {
				fprintf(stdout, "Invalid command.\n");
				continue;
			}
			is_recording = 0;
			is_playback = 1;
			is_learning = 1;
			cur_instrument = 0;
			changeInstrument(MAX_CH-1, 0);
			break;	
		} else if (strcmp(cmd, "erase") == 0){
			is_recording = 0;
			is_playback = 0;
			is_learning = 0;
			resetAll();	
			fprintf(stdout, "Recorded tracks erased\n");
		} else if (strcmp(cmd, "send") == 0){
			fprintf(stdout, "Sending song\n");
			is_recording = 0;
			is_playback = 0;
			is_learning = 0;
			use_ir = 1;
			ir_comm_mode = 1;
			is_sender = 1;
			break;
		} else if (strcmp(cmd, "receive") == 0){
			fprintf(stdout, "Receiving song\n");
			is_recording = 0;
			is_playback = 0;
			is_learning = 0;
			use_ir = 1;
			ir_comm_mode = 1;
			is_sender = 0;
			break;
		} else {
			fprintf(stdout, "Invalid command.\n");
		}
    }
	sei(); //Reenable interrupts

    //Set state based on user input
    
	if (is_recording){
		beginRecording();
	} if (is_playback){
		beginPlayback();
	}

	if(use_ir){
		clear_buffers();
		ir_rx_ready = 0;
		ir_tx_ready = 0;
		if (is_sender){
			state = SEND_SYN;
			sendTask();
		} else {
			state = RECEIVE_SYN;
			receiveTask();
		}	
	}
	ir_comm_mode = 0;
	cleanup_ir();
}

int main(void){
   	// make B.3 an output for PWM
   	DDRB = (1<<PINB3);

	//Make B.2 trigger external interrupts
	PORTB = (1<<INT2);
	EIMSK = (1<<INT2);
   
   	//init the UART -- uart_init() is in uart.c
   	uart_init();
   	stdout = stdin = stderr = &uart_str;
   	printf("Engine initialized...\n\r");

	initMidi();             	//Configures MIDI
    midi_stream_idx = 0;        //Start new midi_stream
    initDDS();              	//Configures DDS
	initTables();               //Initialize sine tables
	initParameters();			//Init FM instruments

   	// init the time counter
	clock_ticks = CLOCK_PD;

   	// timer 0 runs at full rate
   	TCCR0B = 1 ;  
   	//turn off timer 0 overflow ISR
   	TIMSK0 = 0 ;
   	// turn on PWM
  	// turn on fast PWM and OC0A output
  	// at full clock rate, toggle OC0A (pin B3) 
  	// 16 microsec per PWM cycle sample time
  	TCCR0A = (1<<COM0A0) | (1<<COM0A1) | (1<<WGM00) | (1<<WGM01) ; 
 	OCR0A = 128 ; // set PWM to half full scale
	
	// timer 1 ticks at 8000 Hz or 125 microsecs period=2000 ticks
	OCR1A = 1999 ; // 2000 ticks
	TIMSK1 = (1<<OCIE1A) ;
	TCCR1B = 0x09; 	//full speed; clear-on-match
 	TCCR1A = 0x00;	//turn off pwm and oc lines

	//Initialize all instruments to zero at beginning	
	for (uint8_t i = 0; i < MAX_CH; i++){
		changeInstrument(i, 0);
	}

    //IR Communication Setup
    
	OCR2A = 142 ; //143 cycles/half-period
	//  count at full rate
	TCCR2B =  1;	
	// set to toggle OC2A, clear on match,
	TCCR2A = (1<<COM2A0) | (1<<WGM21) ;
	// PORT D.7 is OC2A
	DDRD = (1<<PIND7) ;	
	//********************
  
	// turn on all ISRs
    sei() ;

    //Set default key mapping
	setSampleNotes();
    
    //Enter CLI for initial use
	serialComm();

	if (is_recording){
		beginRecording();
	} if (is_playback){
		beginPlayback();
	}
	
	clearSMPTE();
    while(1){
        while (is_playback && (pb_stream_idx >= 0) && !finished_playback) {
            //Process playback

            //Process the current quarter frame of the prerecorded stream
            play_next_QFM = 0;
            pb_stream_idx = processQFM(pb_stream, pb_stream_idx, pb_stream_len - pb_stream_idx);
            //Wait for ISR to indicate the next quarter frame
			while (is_playback && !play_next_QFM && !finished_recording && !finished_playback){
				updateKeypad(); //Read keypad
			}
		}

		while (is_recording && !finished_recording){
			updateKeypad(); //Read keypad
		}

		if (is_learning){
            //Give user prompt to start playback
			fprintf(stdout, "Now you try\n");

			//Give the user a 3 second countdown before recording starts
			uint8_t second = SMPTE_time->sec;
			while (second == SMPTE_time->sec);
			fprintf(stdout, "3...");
			second = SMPTE_time->sec;
			while (second == SMPTE_time->sec);
			fprintf(stdout, " 2...");
			second = SMPTE_time->sec;
			while (second == SMPTE_time->sec);
			fprintf(stdout, " 1...");
			second = SMPTE_time->sec;
			while (second == SMPTE_time->sec);
			fprintf(stdout, "\nGo!\n");

			//Prepare for recording
			clearSMPTE();
			midi_stream_idx = 0;

            //Record user input for teacher mode
			beginRecording();
			while (!finished_recording){
				updateKeypad();
			}

			//Clean up state
			is_recording = 0;
			finished_recording = 0;
			is_learning = 0;
			finalizeMidiStream();
			clearSMPTE();

			//Calculate and display the user's score using teaching algorithm
			uint16_t corr, incorr;
			calcScore(&corr, &incorr, pb_stream, 
				MIDI_STREAM_OVHD + getTrackLen(pb_stream), MIDI_STREAM_OVHD + getTrackLen(midi_stream));
			uint16_t corr_percent = (100 * corr) / (corr + incorr);
			printf("\n");
			printf("Your score is %i out of 100\n", corr_percent);
			//Final clean up of state
			resetAll();

		} else if (finished_recording){
            //Merge user recording into existing recorded stream and
            //clean up state
			clearSMPTE();
			cur_chan++;
			finished_recording = 0;
			finalizeMidiStream();
			if (is_recording && validHeaderChunk(pb_stream)){
				mergeMidiStreams(pb_stream);
			} 
			
			saveRecording();

			//Turn off real time playback
			buttons = 0;
			playButtonPress();

		}
        
        //Prompt new input
		serialComm();
		
    }

	return 0;
}

ISR (TIMER1_COMPA_vect){ //Fs = 8000
    //Generate 1 msec avg time base for MAIN	
	clock_ticks--;

	if (clock_ticks == 0){
		//Timebase logic
		clock_ticks = CLOCK_PD;
		ir_time++;
		if (ir_comm_mode) {
		//**********************
			// send an ir char if tx is ready and still char in buffer to send
			// and USART is ready
			if ( ir_tx_ready ){ //&& ir_tx_buffer[ir_tx_count]>0
				if (UCSR1A & (1<<UDRE1)) UDR1 = ir_tx_buffer[ir_tx_count++];
				if (ir_tx_count >= buffer_size) ir_tx_ready = 0; // buffer overrun
			}
			
			//**********************
			// recv an ir char if data ready 
			// append character to the received string
			if ( UCSR1A & (1<<RXC1) ) {
				uint8_t c = UDR1;
				if (ir_rx_count != 0 || c == start_token){
					ir_rx_buffer[ir_rx_count++] = c;

					if (ir_rx_count >= buffer_size) {
						ir_rx_ready = 1;
						ir_rx_buffer[buffer_size-1] = 0x00;
						ir_rx_count = buffer_size -1 ; //???
					}
				}
			}
		}
		else {	
			incrementSMPTE(); //Increment global timer
			if (isQuarterFrame()){ //Occurs every 10 msec
				if (is_recording){
					addQuarterFrameMsg();
				} 
				if (is_playback){
					play_next_QFM = 1; //Flag for playback inside main loop
				}
			} 

			// Debounce button presses
			debounce_msec--;
			if (debounce_msec == 0){
				debounce_msec = DEBOUNCE_PD_MSEC;
				debounce_flag = 1;
			}			
			
			//Place 15 second max on all recording and playback due to SRAM capacity
			if (SMPTE_time->sec >= 15){
				printf("Recording finished!\n");
				finished_recording = 1;
			}
		}
	}	
	
	if (!ir_comm_mode) {
		
		//FM synthesis code
		clock_ticks_fm--; //Decrement global FM synthesis time counter
		if (clock_ticks_fm == 0) {
			clock_ticks_fm = CLOCK_FOR_FM;
			for (uint16_t i = 0; i < MAX_CH; i++){
				for (uint16_t j = 0; j < MAX_NOTES; j++){
					// compute exponential decay of FM depth of modulation
					amp_fm1[i][j] = amp_fm1[i][j] - (amp_fm1[i][j] >> decay_fm1[i][j]) ;
				}
			}		
		}

		for (uint16_t i = 0; i < MAX_CH; i++){
			for (uint16_t j = 0; j < MAX_NOTES; j++){
				if (channel_on[i][j]){
				
					amp_fm1[i][j] = amp_fm1[i][j] - (amp_fm1[i][j] >> decay_fm1[i][j]) ;
					// Init the synth
					if (pluck[i][j]==1) {
						amp_fm1[i][j] = max_amp;
						// phase lock the synth
						acc_fm1[i][j] = 0 ;
						acc_main[i][j] = 0;
						pluck[i][j] = 0;
					}

					//the FM DDR -- feeds into final DDR
					acc_fm1[i][j] = acc_fm1[i][j] + inc_fm1[i][j];
					high_fm1[i][j] = (uint8_t)(acc_fm1[i][j] >> 8) ;
					fm1[i][j] = waveTable[sel_fm1[i][j]][high_fm1[i][j]] ;

					//the final output DDR 
					// phase accum = main_DDR_freq + FM_DDR * (FM amplitude)
					acc_main[i][j] = acc_main[i][j] + (inc_main[i][j] + (fm1[i][j]*(amp_fm1[i][j]>>depth_fm1[i][j]))) ;
					high_main[i][j] = (uint8_t)(acc_main[i][j] >> 8) ;
				}
			}
		}
		
		// output the wavefrom sample
		// scale amplitude to use only high byte and shift into range
		// 0 to 255
		int16_t sum= 0;
		for (uint16_t i = 0; i < MAX_CH; i++){
			for (uint16_t j = 0; j < MAX_NOTES; j++){
				if (channel_on[i][j]){
					sum += (((uint16_t) waveTable[sel_main[i][j]][high_main[i][j]])>>7);
				}
			}
		}


		OCR0A = 128 + (sum >> _num_channels_shamt); //Take arithmetic avg of sum
	}
}

ISR (INT2_vect) {
    //Interrupt for entering serialComm during another mode
    //Sets all state and flags required to exit any inner loops in the main loop
	if (is_recording){
		finished_recording = 1;
	}
	if (is_playback){
		finished_playback = 1;
		//Reset DDS
		buttons = 0;
		playButtonPress();
	}
	is_learning = 0;
	cleanup_ir();
	ir_comm_mode = 0;
}
