//
// Created by Albert Liao on 11/19/2013.
//
// Uses a variant of TCP/IP Protocol.
// ses code from ECE 4760 @ Cornell Lab 3.

#define F_CPU 16000000
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <math.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <util/delay.h>

#include "bit_vector.c"

//allows simultaneous send and receive
uint8_t duplex = 0;

//timeout counter  
// time to give up and assume you will never get a response   
#define ir_rx_timeout 1000L //milliseconds
#define ir_tx_timeout 1000L

// string buffers and ISR comm buffers
#define payload_size 64
#define buffer_size 69
volatile char ir_tx_buffer[buffer_size], ir_rx_buffer[buffer_size] ;
// count for filling(sending) buffers
volatile char ir_tx_count, ir_rx_count ;
// read flags
volatile char ir_tx_ready, ir_rx_ready ;
// start and end characters for packet
#define start_token '#'
#define end_token '%'

#define SEND_SYN 0
#define RECEIVE_SYN_ACK 1
#define SEND_PACKET 2
#define RECEIVE_PACKET_ACK 3
#define SEND_FIN 4
#define RECEIVE_FIN_ACK 5
uint8_t state; //Holds FSM state


#define RECEIVE_SYN 6
#define SEND_SYN_ACK 7
#define RECEIVE_PACKET 8
#define SEND_PACKET_ACK 9
#define RECEIVE_FIN 10
#define SEND_FIN_ACK 11

#define DONE 12

#define CORRECT_RESPONSE 6
#define ERROR_RESPONSE 1
#define TIMEOUT_RESPONSE 2
#define RESET_RESPONSE 3 // something REALLY bad happened; start over entirely.
#define DO_NOTHING_RESPONSE 4 // return this if not ready
#define BUFFER_OVERRUN_RESPONSE 5
#define NO_DATA_RESPONSE 7

#define SYN_ID '1'
#define SYN_ACK_ID 'D'
#define PACKET_ACK_ID 'Z'
#define FIN_ID 'j'
#define FIN_ACK_ID 'z'

uint8_t packet_id = 1; // For recieveTask, the id we expect. For sendtask, the one we're sending.
uint8_t received_packet_id = 1; // For receiveTask. The packet_id we just received.
uint8_t last_packet_id = 0;
volatile unsigned long tx_send_time, rx_rec_time;

volatile uint32_t ir_time = 0; // in msec.
uint32_t last_sent_packet_time = 0;
volatile uint8_t use_ir = 0;
uint8_t is_sender = 1;

// pb_stream --> send from here.
// midi_stream --> receive into here.

void clear_buffers(void){
	for (uint8_t i = 0; i < buffer_size; i++){
		ir_tx_buffer[i] = 0;
		ir_rx_buffer[i] = 0;
	}
}

void cleanup_ir(void){
	clear_buffers();
	ir_tx_count = 0;
	ir_rx_count = 0;
	use_ir = 0;
	while ( UCSR1A & (1<<RXC1) ) {
		char dummy =  UDR1;
	}
	UCSR1B = 0;
}

uint8_t is_tx_state(void){
	//Returns 1 iff the FSM is in a tx state, else returns a 0, indicating
	//a rx state
	return  (state == SEND_SYN) ||
			(state == SEND_PACKET) ||
			(state == SEND_FIN) ||
			(state == SEND_SYN_ACK) ||
			(state == SEND_PACKET_ACK) ||
			(state == SEND_FIN_ACK);
}

// IR send
// Input transmitter id and string packet payload
void ir_send_packet(char tx_id, uint8_t ir_data[], uint8_t ir_data_len) {
	char ir_tx_ck_sum, ir_tx_ck_sum1, ir_tx_ck_sum2 ;
	uint8_t i ;
	
	// compute check sum on input data string only
	ir_tx_ck_sum = 0;
	for (i=0; i < ir_data_len; i++)
		ir_tx_ck_sum ^= ir_data[i] ;
	// chop the check sum into two 1/2 bytes and add a bit
	ir_tx_ck_sum1 = 0x10 + (ir_tx_ck_sum & 0x0f) ;
	ir_tx_ck_sum2 = 0x10 + ((ir_tx_ck_sum>>4) & 0x0f) ;
	
	// format send buffer
	ir_tx_buffer[0] = start_token;
	ir_tx_buffer[1] = tx_id;

	//Copy the ir_data to the transmission buffer
	for (uint8_t i = 0; i < ir_data_len; i++){
		ir_tx_buffer[i + 2] = ir_data[i];
	}

	//Add the checksums and end token
	ir_tx_buffer[2 + payload_size]     = ir_tx_ck_sum1;
	ir_tx_buffer[2 + payload_size + 1] = ir_tx_ck_sum2;
	ir_tx_buffer[2 + payload_size + 2] = end_token;
	
  	ir_tx_count = 0 ;
  	ir_tx_ready = 1 ;
	
	tx_send_time = ir_time;
	// wait 
	while (ir_tx_ready && (ir_time < tx_send_time + ir_tx_timeout)) {};
}

//**********************************************************
// IR Receive
// input expected transmitter id
// returns zero if payload is valid and returns payload
// 1 means no data; 2 means buffer overrun; 3 means bad tx id; 4 means bad checksum
char ir_rec_packet(char tx_id, uint8_t ir_data[], uint8_t ir_data_len) {
	char rx_status = 0;
	char ir_rx_ck_sum, ir_rx_ck_sum_ref;
	uint8_t i ;

	// error check

	if (ir_rx_ready == 2) {rx_status = 2; return rx_status;} // buffer overrun
	if (ir_rx_ready == 0) {rx_status = 5; return rx_status;} // no data -- timeout
	if (ir_rx_buffer[1] != tx_id) {rx_status = 3; return rx_status;} // bad transmitter id
	if (ir_rx_ready != 1) {rx_status = 1; return rx_status;} // invalid data
	
	// compute receive checksum
	ir_rx_ck_sum = 0 ;
	for (i=2; i < ir_data_len + 2; i++)
		ir_rx_ck_sum ^= ir_rx_buffer[i] ;
	
	ir_rx_ck_sum_ref = (ir_rx_buffer[buffer_size-3] & 0x0f) +
		 (ir_rx_buffer[buffer_size-2]<<4) ;

    if (ir_rx_ck_sum_ref != ir_rx_ck_sum) 
		{rx_status = 4; return rx_status;}  // bad check sum

	// set up the valid data return stuff
	ir_rx_ready = 0 ;
	ir_rx_count = 0;
	
	memcpy(ir_data, ir_rx_buffer+2, payload_size);

	return rx_status ;
}

// Finish this.
// NOTE! The midi_stream can be a maximum of ~15kB.
uint8_t get_num_packets_necessary(volatile uint8_t* stream) {	
	uint32_t total_length_bits = (getTrackLen(stream) + MIDI_STREAM_OVHD) << 3;
	uint32_t payload_in_bits = payload_size * 7;
	return (uint8_t) ((total_length_bits + payload_in_bits - 1) / payload_in_bits);
}

/** SYN command will be the number of packets. */
void send_syn(void) {
    uint8_t ir_tx_data[2];
	sprintf((void *) ir_tx_data, "%c", get_num_packets_necessary(midi_stream));
	last_packet_id = get_num_packets_necessary(midi_stream);
	// ID is SYN_ID.
    ir_send_packet(SYN_ID, ir_tx_data, 1);
	last_sent_packet_time = ir_time;
}

uint8_t receive_syn(uint8_t ir_rx_data[]) {
    //Called by receiver asynchronously to receive syn commands
    //If a request was received, returns whether it was a proper SYN ('r').
    
	// ID is SYN_ID.
	uint8_t rx_status = ir_rec_packet(SYN_ID, ir_rx_data, 1);
    if (rx_status == 0){
		last_packet_id = (uint8_t) ir_rx_data[0];
        return CORRECT_RESPONSE;
    } else if (rx_status == 5) {
		if (ir_time - last_sent_packet_time > ir_rx_timeout) {
			return TIMEOUT_RESPONSE;
		}
		else {
			return DO_NOTHING_RESPONSE;
		}
	} else if (rx_status == 2) {
        return BUFFER_OVERRUN_RESPONSE;
    } else {
        return ERROR_RESPONSE;
    }
	
	// Can add more granularity if needed.
}

void send_syn_ack(void) {
    uint8_t ir_tx_data[1] = {'z'};
	// ID is SYN__ACK_ID.
    ir_send_packet(SYN_ACK_ID, ir_tx_data, 1);
	last_sent_packet_time = ir_time;
}

uint8_t receive_syn_ack(uint8_t ir_rx_data[]) {
    //Called by receiver asynchronously to receive syn_ack commands
    //If a request was received, returns whether it was a proper SYN_ACK ('z').
    
	// ID is SYN_ACK_ID.
	uint8_t rx_status = ir_rec_packet(SYN_ACK_ID, ir_rx_data, 1);
    if (rx_status == 0){
		return ir_rx_data[0] == 'z' ? CORRECT_RESPONSE : ERROR_RESPONSE;
    } else if (rx_status == 5) {
		if (ir_time - last_sent_packet_time > ir_rx_timeout) {
			return TIMEOUT_RESPONSE;
		}
		else {
			return DO_NOTHING_RESPONSE;
		}
	} else if (rx_status == 2) {
        return BUFFER_OVERRUN_RESPONSE;
    } else {
        return ERROR_RESPONSE;
    }
	// Can add more granularity if needed.
}

void write_to_tx(uint8_t ir_tx_data[]){
	uint32_t midi_bit_idx = ((uint16_t) packet_id - 1) * (payload_size * 7);

	//Write data to the transmission buffer
	for (uint16_t i = 0; i < payload_size; i++){
		ir_tx_data[i] = get7(midi_stream, midi_bit_idx);
		midi_bit_idx += 7;
	}
}

void read_from_rx(uint8_t ir_rx_data[]){
	//Save data from the receive buffer
	uint32_t midi_bit_idx = ((uint16_t) packet_id - 1) * (payload_size * 7);
	for (uint16_t i = 0; i < payload_size; i++){
		set7(pb_stream, midi_bit_idx, ir_rx_data[i]);
		midi_bit_idx += 7;
	}
}
void send_packet(void) {
	uint8_t ir_tx_data[payload_size + 1];
	write_to_tx(ir_tx_data);
    ir_send_packet(packet_id, ir_tx_data, payload_size);
	last_sent_packet_time = ir_time;
}

uint8_t receive_packet(uint8_t ir_rx_data[]) {
    //Called by receiver asynchronously to receive syn_ack commands
    //If a request was received, returns whether it was a proper SYN_ACK ('z').
    
	// ID is packet_id.
	uint8_t rx_status = ir_rec_packet(packet_id, ir_rx_data, payload_size);
    if (rx_status == 0){
		read_from_rx(ir_rx_data);
		return CORRECT_RESPONSE;

    } else if (rx_status == 5) {
		if (ir_time - last_sent_packet_time > ir_rx_timeout) {
			return TIMEOUT_RESPONSE;
		}
		else {
			return DO_NOTHING_RESPONSE;
		}
	} else if (rx_status == 2) {
        return BUFFER_OVERRUN_RESPONSE;
    } else {
        return ERROR_RESPONSE;
    }
	// Can add more granularity if needed.	
}

void send_packet_ack(void) {
    uint8_t ir_tx_data[1] = {'p'};

    ir_send_packet(PACKET_ACK_ID, ir_tx_data, 1);
	last_sent_packet_time = ir_time;
}

uint8_t receive_packet_ack(uint8_t ir_rx_data[]) {
    //Called by receiver asynchronously to receive syn_ack commands
    //If a request was received, returns whether it was a proper SYN_ACK ('z').
    
	// ID is PACKET_ACK_ID.
	uint8_t rx_status = ir_rec_packet(PACKET_ACK_ID, ir_rx_data, 1);
    if (rx_status == 0){
        return ir_rx_data[0] == 'p' ? CORRECT_RESPONSE : ERROR_RESPONSE;
    } else if (rx_status == 5) {
		if (ir_time - last_sent_packet_time > ir_rx_timeout) {
			return TIMEOUT_RESPONSE;
		}
		else {
			return DO_NOTHING_RESPONSE;
		}		
	} else if (rx_status == 2) {
        return BUFFER_OVERRUN_RESPONSE;
    } else {
        return ERROR_RESPONSE;
    }
	// Can add more granularity if needed.	
}

void send_fin(void) {
    uint8_t ir_tx_data[1] = {'f'};
    ir_send_packet(FIN_ID, ir_tx_data, 1);
	last_sent_packet_time = ir_time;
}

uint8_t receive_fin(uint8_t ir_rx_data[]) {
    //Called by receiver asynchronously to receive fin commands
    //If a request was received, returns whether it was a proper FIN ('z').
    
	// ID is FIN_ID.
	uint8_t rx_status = ir_rec_packet(FIN_ID, ir_rx_data, 1);
    if (rx_status == 0){
        return ir_rx_data[0] == 'f' ? CORRECT_RESPONSE : ERROR_RESPONSE;
    } else if (rx_status == 5) {
		if (ir_time - last_sent_packet_time > ir_rx_timeout) {
			return TIMEOUT_RESPONSE;
		}
		else {
			return DO_NOTHING_RESPONSE;
		}		
	} else if (rx_status == 2) {
        return BUFFER_OVERRUN_RESPONSE;
    } else {
        return ERROR_RESPONSE;
    }
	// Can add more granularity if needed.
}

void send_fin_ack(void) {
    uint8_t ir_tx_data[1] = {'q'};
    ir_send_packet(FIN_ACK_ID, ir_tx_data, 1);
	last_sent_packet_time = ir_time;
}

uint8_t receive_fin_ack(uint8_t ir_rx_data[]) {
    //Called by receiver asynchronously to receive fin_ack commands
    //If a request was received, returns whether it was a proper FIN_ACK ('q').
    
	// ID is FIN_ACK_ID.
	uint8_t rx_status = ir_rec_packet(FIN_ACK_ID, ir_rx_data, 1);
    if (rx_status == 0){
        return ir_rx_data[0] == 'q' ? CORRECT_RESPONSE : ERROR_RESPONSE;
    } else if (rx_status == 5) {
		if (ir_time - last_sent_packet_time > ir_rx_timeout) {
			return TIMEOUT_RESPONSE;
		}
		else {
			return DO_NOTHING_RESPONSE;
		}		
	} else if (rx_status == 2) {
        return BUFFER_OVERRUN_RESPONSE;
    } else {
        return ERROR_RESPONSE;
    }
	// Can add more granularity if needed.
}

void sendTask(void) {
    //Complete FSM for sending a MIDI stream
    //See http://people.ece.cornell.edu/land/courses/ece4760/FinalProjects/f2013/mjs476_ajl264/mjs476_ajl264/index.html
	uint8_t ir_rx_data[buffer_size]; //Holds the challenge
    state = SEND_SYN;
	packet_id = 1;
	last_sent_packet_time = ir_time;
	last_packet_id = 0;
	uart_init1();

	memcpy(midi_stream, pb_stream, getTrackLen(pb_stream) + MIDI_STREAM_OVHD);

	while (use_ir && is_sender) {
		if (state == SEND_SYN) {
			send_syn();
			// send_syn, MUST SET TIME FOR ALL SENDS.
			// THIS must determine last_packet_id.
			state = RECEIVE_SYN_ACK;
		}
		else if (state == RECEIVE_SYN_ACK) {
			uint8_t status = receive_syn_ack(ir_rx_data);
			if (status == CORRECT_RESPONSE){ // SHOULD BE CORRECT_DATA
				state = SEND_PACKET;
			}
			else if (status == TIMEOUT_RESPONSE || status == ERROR_RESPONSE){
				state = SEND_SYN;
			}
			else if (status == RESET_RESPONSE) { // Bad error.
				state = SEND_SYN;
			}
		}
		else if (state == SEND_PACKET) {
			send_packet();
			printf("Sending packet %i of %i\n", packet_id, last_packet_id);
			state = RECEIVE_PACKET_ACK;
			// In send_packet, must set time.
		}
		else if (state == RECEIVE_PACKET_ACK) {
			uint8_t status = receive_packet_ack(ir_rx_data);
			if (status == CORRECT_RESPONSE) {
				packet_id++;
				if (packet_id > last_packet_id) {
					state = SEND_FIN;
				} else {
					state = SEND_PACKET;
				}
			}
			else if (status == TIMEOUT_RESPONSE || status == ERROR_RESPONSE) {
				if(packet_id == 1){
					 state = SEND_SYN;
				} else {
					state = SEND_PACKET;
				}
			}
			else if (status == RESET_RESPONSE) { // Bad error.
				state = SEND_SYN;
			}
		}
		else if (state == SEND_FIN) {
			send_fin();
			state = RECEIVE_FIN_ACK;
		}
		else if (state == RECEIVE_FIN_ACK) {
			break;

			/*uint8_t status = receive_fin_ack(ir_rx_data);
			if (status == CORRECT_RESPONSE) {
				state = DONE;
			} 
			else if (status == TIMEOUT_RESPONSE || status == ERROR_RESPONSE) {
				state = SEND_FIN;
			}
			else if (status == RESET_RESPONSE) { // Bad error.
				state = SEND_SYN;
			} */
		}
		else if (state == DONE) {
			state = SEND_SYN; // Reset for next time.
			use_ir = 0;
		}
	}

	cleanup_ir();

}

void receiveTask(void) {
    //Complete FSM for receiving a MIDI stream
    //See http://people.ece.cornell.edu/land/courses/ece4760/FinalProjects/f2013/mjs476_ajl264/mjs476_ajl264/index.html
	uint8_t ir_rx_data[buffer_size]; //Holds the challenge
    state = RECEIVE_SYN;
	packet_id = 1;
	last_sent_packet_time = ir_time;
	uint16_t missed_ack_count = 0;
	last_packet_id = 0;
	uart_init1();
	
	while (use_ir && !is_sender) {
		if (state == RECEIVE_SYN) {
			uint8_t status = receive_syn(ir_rx_data);
			// receive_syn MUST get last_packet_id!!!
			if (status == CORRECT_RESPONSE){ // SHOULD BE CORRECT_DATA
				state = SEND_SYN_ACK;
			}
			else { // Error of some sort.
				// Do nothing, since we need to receive a proper syn.
			}		
		}
		else if (state == SEND_SYN_ACK) {
			missed_ack_count = 0;
			send_syn_ack();
			state = RECEIVE_PACKET;
		}
		else if (state == RECEIVE_PACKET) {
			uint8_t status = receive_packet(ir_rx_data);
			if (status == CORRECT_RESPONSE) {
				packet_id++; 
				state = SEND_PACKET_ACK;
				printf("Received packet %i of %i\n", packet_id, last_packet_id);
			}
			else if (status == TIMEOUT_RESPONSE || status == ERROR_RESPONSE) {
				if (missed_ack_count > 5){
					state = SEND_SYN_ACK;
				} else if (status == ERROR_RESPONSE && ir_rx_buffer[1] == SYN_ID){
					missed_ack_count++;
				} else {
					state = SEND_PACKET_ACK;
				}
			}
		}
		else if (state == SEND_PACKET_ACK) {
			send_packet_ack();
			if (packet_id > last_packet_id) { // if num packets = 2, packet_id = 3 after 2nd is sent.
				state = RECEIVE_FIN;
			}
			else {
				state = RECEIVE_PACKET;
			}
		}
		else if (state == RECEIVE_FIN) {
			uint8_t status = receive_fin(ir_rx_data);
			if (status == CORRECT_RESPONSE) {
				state = SEND_FIN_ACK;
			} else if (status == TIMEOUT_RESPONSE || status == ERROR_RESPONSE) {
				state = SEND_PACKET_ACK;
			}
		}
		else if (state == SEND_FIN_ACK) {
			send_fin_ack();
			state = DONE;
		}
		else if (state == DONE) {
			// To ensure that the sender finishes too!!
			break;
		}
	
	}
	cleanup_ir();
}


