//
//  midi_lib.c
//  
//
//  Created by Matt Stern on 10/28/13.
//
//

#include <string.h>
#include <util/atomic.h>

//Important:
//Assumes that Timer0 is operating at full clock rate
//and calls midiISR() every 1 msec

// Set resolution of SMPTE timer
// 25 frame/sec * 40 subframe/frame = 1 subframe/msec
#define FRAME_RATE    25
#define SUBFRAME_RATE 40

//Possible quarter frames that indicate a QF msg
//should be sent (since modulus operator is expensive)
uint8_t QFRAME_RATE;
uint8_t QF0;
uint8_t QF1;
uint8_t QF2;
uint8_t QF3;

//Useful macros
#define LSN(B) (B & 0x0f)        //Least significant nibble of a byte
#define MSN(B) ((B & 0xf0) >> 4) //Most signficant nibble of a byte
#define LSB(W) (W & 0xff)        //Least significant byte of a word
#define MSb(B) (B & 0x80)		 //Most significant bit of a byte

//Struct to hold the current delta time in SMPTE format
typedef struct {
    volatile uint8_t hr;
    volatile uint8_t min;
    volatile uint8_t sec;
    volatile uint8_t f;  //Frame
    volatile uint8_t sf; //Subframe
} SMPTE_time_helper;

//Time variables
volatile SMPTE_time_helper SMPTE_time_obj = {
    .hr  = 0,
    .min = 0,
    .sec = 0,
    .f   = 0,
    .sf  = 0
};
volatile SMPTE_time_helper * SMPTE_time = &SMPTE_time_obj;

//MIDI stream
#define MIDI_STREAM_SIZE 6000
#define MIDI_STREAM_HEADER_LEN 14
#define MIDI_STREAM_OVHD 22 //Number of bytes needed for the header and track chunks
volatile uint8_t  midi_stream[MIDI_STREAM_SIZE];
volatile int32_t  midi_stream_idx = 0;
volatile uint8_t  qframe_idx = 0; //Takes on values 0x00, 0x10, ... 0x70

//Real time playback variables
uint8_t pb_stream[MIDI_STREAM_SIZE];
int32_t pb_stream_len;
int32_t pb_stream_idx;

//Holds the current channel we are operating on
uint8_t num_chan = 1;

//--------------------Helper functions-------------------------------

uint16_t get_uint16(volatile uint8_t* stream, uint32_t idx){
    //Helper to get a 16-bit word in `stream` starting at `idx`
    uint16_t bit0 = (uint16_t)  stream[idx+1];
	uint16_t bit1 = ((uint16_t) stream[idx]  ) << 8;

    return bit1 | bit0;
}

void set_uint16(volatile uint8_t* stream, uint32_t idx, uint16_t val){
    //Helper to set a 16-bit word in `stream` starting at `idx`
    stream[idx]   = (uint8_t) (val >> 8);
    stream[idx+1] = (uint8_t) (val & 0xff);
}

uint32_t get_uint32(volatile uint8_t* stream, uint32_t idx){
    //Helper to get a 32-bit word in `stream` starting at `idx`
	uint32_t bit0 =  (uint32_t)  stream[idx+3];
	uint32_t bit1 = ((uint32_t) (stream[idx+2])) << 8;
	uint32_t bit2 = ((uint32_t) (stream[idx+1])) << 16;
	uint32_t bit3 = ((uint32_t) (stream[idx]  )) << 24;

    return bit3 | bit2 | bit1 | bit0;
}

void set_uint32(volatile uint8_t* stream, uint32_t idx, uint32_t val){
    //Helper to set a 32-bit word in `stream` starting at `idx`
    stream[idx]   = (uint8_t) (val >> 24);
    stream[idx+1] = (uint8_t) ((val >> 16) & 0xff);
    stream[idx+2] = (uint8_t) ((val >> 8)  & 0xff);
    stream[idx+3] = (uint8_t) (val & 0xff);
} 

uint8_t validHeaderChunk(volatile uint8_t* stream){
	//Returns 1 iff the MIDI header for this stream is valid for the 
	//purposes of this library
	uint8_t passed = 1;

	//Add chunk ID, which is the character string "MThd"
    passed &= stream[0] == 0x4d; //'M'
    passed &= stream[1] == 0x54; //'T'
    passed &= stream[2] == 0x68; //'h'
    passed &= stream[3] == 0x64; //'d'
    
    //Add the size of the header chunk in bytes, which is 6 by default
    passed &= stream[4] == 0x00;
    passed &= stream[5] == 0x00;
    passed &= stream[6] == 0x00;
    passed &= stream[7] == 0x06;

    //Add the format type, which is default 0 to allow a single track only
	//We will be merging all tracks into one single track for simplicity
    passed &= stream[8] == 0x00;
    passed &= stream[9] == 0x00;
    
    //Add the number of tracks, which we set to 1 by default
    passed &= stream[10] == 0x00;
    passed &= stream[11] == 0x01;
    
    //Add the time division, which will be in frames per second mode
	
	uint8_t neg_frame_rate = ~(FRAME_RATE) + 1; //We lose negative sign in uint8_t array
    passed &= stream[12] == neg_frame_rate; 
    passed &= stream[13] == SUBFRAME_RATE;

	return passed;
}

uint8_t validTrackChunk(volatile uint8_t* stream){
	//Validates the _first_ track chunk in the stream
	uint8_t passed = 1;

    //Add chunk ID, which is the character string "MTrk"
    passed &= stream[14]   == 0x4d; //'M'
    passed &= stream[15] == 0x54; //'T'
    passed &= stream[16] == 0x72; //'r'
    passed &= stream[17] == 0x6b; //'k'

	return passed;
}

uint32_t getTrackLen(volatile uint8_t* new_stream){
	//Returns the track length according to the _first_ track chunk header.
	return get_uint32(new_stream, MIDI_STREAM_HEADER_LEN + 4);
}

int32_t indexOfQFM(volatile uint8_t* stream, int32_t start, int32_t num){
    //Searches `num` bytes in `stream` starting for 0xf1.
    //Returns the idx of the first occurrence >= idx `start, or
    //-1 if we do not find 0xf1
	if (num < 0){
		return -1;
	}
    uint8_t* ptr = (uint8_t *) memchr((void *) &stream[start], 0xf1, num);
    return ptr == NULL ? -1 : ptr - stream;
}
                   
void cpyInc(volatile uint8_t* dest, volatile uint8_t* src, int32_t* d_idx_ptr, int32_t* s_idx_ptr, uint32_t amt){
	//Copies `amt` bytes starting at dest[*d_idx] to src[*s_idx] and increments
	//*d_idx and *s_idx by `amt`
	memcpy((void *) &dest[*d_idx_ptr], (void *) &src[*s_idx_ptr], amt);
	*d_idx_ptr += amt;
	*s_idx_ptr += amt;
}

void cpyQFM(volatile uint8_t* dest, volatile uint8_t* src, int32_t* d_idx_ptr, int32_t* s_idx_ptr, int32_t* oth_idx_ptr){
	//Copies 2 bytes from dest to src as in cpyInc, but also increments a third idx by 2
    //for convenience
	cpyInc(dest, src, d_idx_ptr, s_idx_ptr, 2);
	*oth_idx_ptr += 2;
}

//--------------------Core functions-----------------------------------

void incrementSMPTE(void){
    //Increments the global SMPTE_time struct, whose resolution
    //is 1 msec per subframe
    SMPTE_time->sf += 1;
    if (SMPTE_time->sf >= SUBFRAME_RATE){
        SMPTE_time->sf = 0;
        SMPTE_time->f += 1;
    } else { return; }
    
    if (SMPTE_time->f >= FRAME_RATE){
        SMPTE_time->f = 0;
        SMPTE_time->sec += 1;
    } else { return; }
    
    if (SMPTE_time->sec >= 60){
        SMPTE_time->sec = 0;
        SMPTE_time->min += 1;
    } else { return; }
    
    if (SMPTE_time->min >= 60){
        SMPTE_time->min = 0;
        SMPTE_time->hr += 1;
    } else { return; }
}

void clearSMPTE(void){
    //Clear SMPTE_time
    SMPTE_time->hr  = 0;
    SMPTE_time->min = 0;
    SMPTE_time->sec = 0;
    SMPTE_time->f   = 0;
    SMPTE_time->sf  = 0;
}

uint8_t isQuarterFrame(void){
    //Returns true iff QFRAME_RATE % SMPTE->sf == 0
    //Will be faster than using a modulus operator
    uint8_t subframe = SMPTE_time->sf;
    return (subframe == QF0) || (subframe == QF1) ||
            (subframe == QF2) || (subframe == QF3);
    
}

void addQuarterFrameMsg(void){
    //Adds a quarter frame message to `midi_stream`
    //See http://www.blitter.com/~russtopia/MIDI/~jglatt/tech/mtc.htm
    midi_stream[midi_stream_idx++] = 0xf1; //Quarter frame status byte
    
    switch (qframe_idx) {
        case 0x00:
            midi_stream[midi_stream_idx++] = qframe_idx | LSN(SMPTE_time->f);
            break;
        case 0x10:
            midi_stream[midi_stream_idx++] = qframe_idx | MSN(SMPTE_time->f);
            break;
        case 0x20:
            midi_stream[midi_stream_idx++] = qframe_idx | LSN(SMPTE_time->sec);
            break;
        case 0x30:
            midi_stream[midi_stream_idx++] = qframe_idx | MSN(SMPTE_time->sec);
            break;
        case 0x40:
            midi_stream[midi_stream_idx++] = qframe_idx | LSN(SMPTE_time->min);
            break;
        case 0x50:
            midi_stream[midi_stream_idx++] = qframe_idx | MSN(SMPTE_time->min);
            break;
        case 0x60:
            midi_stream[midi_stream_idx++] = qframe_idx | LSN(SMPTE_time->hr);
            break;
        case 0x70:
            midi_stream[midi_stream_idx++] = qframe_idx | MSN(SMPTE_time->hr);
            break;
        default:
            break;
    }
    
    qframe_idx = qframe_idx == 0x70 ? 0x00 : qframe_idx + 0x10;
}

void initMidi(void){
	//set global variables on initialization
	QFRAME_RATE = SUBFRAME_RATE >> 2;
	QF0 = 0;
	QF1 = QFRAME_RATE;
	QF2 = QFRAME_RATE * 2;
	QF3 = QFRAME_RATE * 3;
}

void addHeaderChunk(void){
    //Add header chunk to the MIDI stream
    //See: http://www.sonicspot.com/guide/midifiles.html
    //Note that MIDI words are Big Endian
    //The length of the header is 14 bytes
    
    //Add chunk ID, which is the character string "MThd"
    midi_stream[midi_stream_idx++] = 0x4d; //'M'
    midi_stream[midi_stream_idx++] = 0x54; //'T'
    midi_stream[midi_stream_idx++] = 0x68; //'h'
    midi_stream[midi_stream_idx++] = 0x64; //'d'
    
    //Add the size of the header chunk in bytes, which is 6 by default
    midi_stream[midi_stream_idx++] = 0x00;
    midi_stream[midi_stream_idx++] = 0x00;
    midi_stream[midi_stream_idx++] = 0x00;
    midi_stream[midi_stream_idx++] = 0x06;
    
    //Add the format type, which is default 0 to allow a single track only
	//We will be merging all tracks into one single track for simplicity
    midi_stream[midi_stream_idx++] = 0x00;
    midi_stream[midi_stream_idx++] = 0x00;
    
    //Add the number of tracks, which we set to 1 by default
    midi_stream[midi_stream_idx++] = 0x00;
    midi_stream[midi_stream_idx++] = 0x01;
    
    //Add the time division, which will be in frames per second mode
    midi_stream[midi_stream_idx++] = -FRAME_RATE;
    midi_stream[midi_stream_idx++] = SUBFRAME_RATE;
}

void addTrackChunk(uint32_t midi_stream_len){
    //Add track chunk to the MIDI stream
    //See: http://www.sonicspot.com/guide/midifiles.html
    //Note that MIDI words are Big Endian
    //The length of the track chunk is 8 bytes + the length of the
    //track that follows
    
    //Add chunk ID, which is the character string "MTrk"
    midi_stream[midi_stream_idx++] = 0x4d; //'M'
    midi_stream[midi_stream_idx++] = 0x54; //'T'
    midi_stream[midi_stream_idx++] = 0x72; //'r'
    midi_stream[midi_stream_idx++] = 0x6b; //'k'
    
    //Add chunk size
    set_uint32(midi_stream, midi_stream_idx, midi_stream_len);
    midi_stream_idx += 4;
}

void createMidiStream(void){
    //Overwrites the existing MIDI stream and starts a new MIDI stream
    midi_stream_idx = 0;
}

void finalizeMidiStream(){
    //Add the header and track chunks to the MIDI stream, making it a
    //valid MIDI stream to be played back and transmitted
    
    uint32_t midi_stream_len = midi_stream_idx;
    //Move bytes over to make room for header and track chunks
    for (int32_t i = midi_stream_len - 1; i >= 0; i--){
        midi_stream[i + MIDI_STREAM_OVHD] = midi_stream[i];
    }
    
    //Add header and track chunks
    midi_stream_idx = 0;
    addHeaderChunk();
    addTrackChunk(midi_stream_len);
    
    //Set index to the next unused value in the array
    midi_stream_idx = midi_stream_len + MIDI_STREAM_OVHD;
}

uint8_t mergeMidiStreams(volatile uint8_t new_stream[]){
    //Merge the _one track_ `new_stream` into the current _one track_
	//`midi_stream` data between the corresponding quarter frame messages 
	//will be put in between those same messages in the output stream
	//Precondition: Both tracks have the same SMPTE time division
	//_and_ they both start at SMPTE time 00:00:00:00:00.
	//Returns 1 iff the merge was successful.

	if (!validHeaderChunk(midi_stream) || !validHeaderChunk(new_stream)){
		return 0;
	}
	if (!validTrackChunk(midi_stream) || !validTrackChunk(new_stream)){
		return 0;
	}

	uint32_t m_len = getTrackLen(midi_stream) + MIDI_STREAM_OVHD;
	uint32_t n_len = getTrackLen(new_stream)  + MIDI_STREAM_OVHD;

	volatile uint8_t output_stream[MIDI_STREAM_SIZE];
    int32_t o_idx = 0;
    
	//Create "current" idxs for `midi_stream` and `new_stream`
    //Invariant: At the start of each while loop iteration, this index
    //is at the idx directly _after_ a QFM
	int32_t m_idx_cur = MIDI_STREAM_OVHD;
	int32_t n_idx_cur = MIDI_STREAM_OVHD;
    
    //Copy the first QFM from `midi_stream` to `output_stream`
	cpyQFM(output_stream, midi_stream, &o_idx, &m_idx_cur, &n_idx_cur);
    
    //Create "scout idxs" for `midi_stream` and `output_stream`
    //These will mark the idxs of the next QFM in the respective streams
    int32_t m_idx_sct;
    int32_t n_idx_sct;
    
	while (m_idx_cur < m_len - 1 && n_idx_cur < n_len - 1){
        //Find the next QFM in each stream
        m_idx_sct = indexOfQFM(midi_stream, m_idx_cur, m_len - m_idx_cur);
        n_idx_sct = indexOfQFM(new_stream,  n_idx_cur, n_len - n_idx_cur);

		if (m_idx_sct != -1 && n_idx_sct != -1){
			//Normal case, there are status bytes in both streams
			cpyInc(output_stream, midi_stream, &o_idx, &m_idx_cur, m_idx_sct - m_idx_cur);
			cpyInc(output_stream, new_stream,  &o_idx, &n_idx_cur, n_idx_sct - n_idx_cur);
			cpyQFM(output_stream, midi_stream, &o_idx, &m_idx_cur, &n_idx_cur);
		} else if (m_idx_sct == -1){
			//There are no more status bytes left in `midi_stream`, so copy the rest over
			cpyInc(output_stream, midi_stream, &o_idx, &m_idx_cur, m_len - m_idx_cur);
		} else if (n_idx_sct == -1){
			//There are no more status bytes left in `new_stream`, so copy the rest
            cpyInc(output_stream, new_stream,  &o_idx, &n_idx_cur, n_len - n_idx_cur);
		}
    }
    
    if (m_idx_cur < m_len - 1){
        //Copy the rest of `new_stream` into `output_stream`
        cpyInc(output_stream, midi_stream, &o_idx, &m_idx_cur, m_len - m_idx_cur);
    } else if (n_idx_cur < n_len - 1){
        //Copy the rest of `midi_stream` into `output_stream`
        cpyInc(output_stream, new_stream,  &o_idx, &n_idx_cur, n_len - n_idx_cur);
    }

	midi_stream_idx = o_idx;
	memcpy((void *) midi_stream, (void *) output_stream, midi_stream_idx);
	finalizeMidiStream();

    
	return 1;
}

void addTimeDelta(uint32_t ticks){
    //Converts an absolute number of 1 msec ticks to a proper MIDI delta time
	//and appends it to `midi_stream`
	//See http://www.recordingblogs.com/sa/tabid/88/Default.aspx?topic=MIDI+event
	if (ticks){
		uint8_t  dt_holder[5];
		uint8_t  dt_holder_head = 5;
		do {
			dt_holder_head--;
			dt_holder[dt_holder_head] = ticks & 0x7f;
			ticks >>= 7;
		} while ( ticks );
	
		//Add proper variable length delta time MSb flags
		for (uint8_t i = dt_holder_head; i < 4; i++){
			dt_holder[i] |= 0x80;
		}

		uint8_t len = 5 - dt_holder_head;
		memcpy(
  	         (void *) &midi_stream[midi_stream_idx],
  	         (void *) &dt_holder[dt_holder_head],
  	         len
  	  );
		midi_stream_idx += len;
	} else {
		midi_stream[midi_stream_idx++] = 0x00;
	}
}


void changeChannel(volatile uint8_t stream[], uint8_t chan_num, size_t max){
	//Sets channel number of the events in `stream` to `chan_num`
	//Note that `num_chan must be 0x0 - 0xf, or the stream 
	//will be corrupted
	int32_t idx = MIDI_STREAM_OVHD;

	while (idx < max){
		if (stream[idx] == 0xf1){
			idx += 2;
		} else {
			stream[idx] = (stream[idx] & 0xf0) | (chan_num);
			idx += 3;
		}
	}
}

void midiEvent(uint8_t event_type, uint8_t channel, uint8_t param_1, uint8_t param_2){
	//Wrapper to add a MIDI event to the current `midi_stream`
	//Operation is atomic so status messages do not break up this event
	//See: http://www.sonicspot.com/guide/midifiles.htm
	//Also, see: http://www.nongnu.org/avr-libc/user-manual/group__util__atomic.html
	//for explanation on atomic blocks with AVR
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE){
		midi_stream[midi_stream_idx++] = (event_type << 4) | channel;
		midi_stream[midi_stream_idx++] = param_1;
		midi_stream[midi_stream_idx++] = param_2;
	}
}
