//
//  midi_lib_test.c
//  
//
//  Created by Matt Stern on 10/30/13.
//
//

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <stdio.h>

#include "midi_lib.c"
#include "uart.h"
#include "uart.c"

// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

//--------------------Helper functions-------------------------------
void reset_stream(){
	qframe_idx = 0;
	midi_stream_idx = 0;
	clearSMPTE();
}

void quarter_frame_tick(){
	for (uint16_t i = 0; i < (FRAME_RATE >> 2); i++){
		incrementSMPTE();
	}
}

uint8_t assert_SMPTE_time(uint8_t hr, uint8_t min, uint8_t sec, uint8_t f, uint8_t sf){
    return (SMPTE_time->hr  == hr)
        && (SMPTE_time->min == min)
        && (SMPTE_time->sec == sec)
        && (SMPTE_time->f   == f  )
        && (SMPTE_time->sf  == sf );
}

void cmp8(uint8_t exp, uint8_t act){
	fprintf(stdout, "Expected: %i, Actual: %i\n", exp, act);
}

void cmp32(uint32_t exp, uint32_t act){
	fprintf(stdout, "Expected: %ld, Actual: %ld\n", exp, act);
}

//--------------------Begin tests-------------------------------------

uint8_t LSN_test(void){
    uint8_t passed = 1;

    passed &= (LSN(0xab) == 0x0b);
    passed &= (LSN(0xef) == 0x0f);
    passed &= (LSN(0x00) == 0x00);
    
    return passed;
}

uint8_t MSN_test(void){
    uint8_t passed = 1;
    
    passed &= (MSN(0xab) == 0x0a);
    passed &= (MSN(0xef) == 0x0e);
    passed &= (MSN(0x00) == 0x00);
    
    return passed;
}

uint8_t get_uint16_test(void){
    uint8_t passed = 1;
    
    volatile uint8_t array[4] = {0x10, 0x21, 0x32, 0xfe};

    passed &= (get_uint16(array, 0) == 0x1021);
    passed &= (get_uint16(array, 1) == 0x2132);
    passed &= (get_uint16(array, 2) == 0x32fe);
    
	array[0] = 0x23;
	array[1] = 0xab;
	array[2] = 0xef;
	array[3] = 0x00;
    passed &= (get_uint16(array, 0) == 0x23ab);
    passed &= (get_uint16(array, 1) == 0xabef);
    passed &= (get_uint16(array, 2) == 0xef00);
    
    return passed;
}

uint8_t get_uint32_test(void){
    uint8_t passed = 1;
    
    volatile uint8_t array[6] = {0x10, 0x21, 0x32, 0xfe, 0x23, 0xab};

    passed &= (get_uint32(array, 0) == 0x102132fe);
	passed &= (get_uint32(array, 1) == 0x2132fe23);
    passed &= (get_uint32(array, 2) == 0x32fe23ab);
    
	array[0] = 0x23;
	array[1] = 0xab;
	array[2] = 0xef;
	array[3] = 0x00;
	array[4] = 0x10;
	array[5] = 0xea;
    passed &= (get_uint32(array, 0) == 0x23abef00);
    passed &= (get_uint32(array, 1) == 0xabef0010);
    passed &= (get_uint32(array, 2) == 0xef0010ea);
    
    return passed;
}

uint8_t set_uint16_test(void){
    uint8_t passed = 1;
    
    volatile uint8_t array[4];
    set_uint16(array, 0, 0x1234);
    passed &= (array[0] == 0x12);
    passed &= (array[1] == 0x34);
    
    set_uint16(array, 1, 0x4567);
    passed &= (array[0] == 0x12);
    passed &= (array[1] == 0x45);
    passed &= (array[2] == 0x67);
    
    set_uint16(array, 2, 0x89ab);
    passed &= (array[0] == 0x12);
    passed &= (array[1] == 0x45);
    passed &= (array[2] == 0x89);
    passed &= (array[3] == 0xab);
    
    return passed;
}

uint8_t set_uint32_test(void){
    uint8_t passed = 1;
    
    volatile uint8_t array[6];
    set_uint32(array, 0, 0x01234567);
    passed &= (array[0] == 0x01);
    passed &= (array[1] == 0x23);
    passed &= (array[2] == 0x45);
    passed &= (array[3] == 0x67);
	
    set_uint32(array, 1, 0x89abcdef);
    passed &= (array[0] == 0x01);
    passed &= (array[1] == 0x89);
    passed &= (array[2] == 0xab);
    passed &= (array[3] == 0xcd);
    passed &= (array[4] == 0xef);
	
    set_uint32(array, 2, 0xedcba987);
    passed &= (array[0] == 0x01);
    passed &= (array[1] == 0x89);
    passed &= (array[2] == 0xed);
    passed &= (array[3] == 0xcb);
    passed &= (array[4] == 0xa9);
    passed &= (array[5] == 0x87);
    
    return passed;
}

uint8_t indexOfQFM_test(void){
	uint8_t passed = 1;

	volatile uint8_t array[12] = 
		{0xaa, 0xbb, 0xf1, 0xcc, 
		 0xaa, 0xbb, 0xf1, 0xcc,
		 0xaa, 0xbb, 0xf1, 0xcc};
	
	passed &= indexOfQFM(array, 0, 12) == 2;
	passed &= indexOfQFM(array, 1, 11) == 2;
	passed &= indexOfQFM(array, 2, 10) == 2;
	passed &= indexOfQFM(array, 3,  9) == 6;
	passed &= indexOfQFM(array, 4,  8) == 6;
	passed &= indexOfQFM(array, 5,  7) == 6;
	passed &= indexOfQFM(array, 6,  6) == 6;
	passed &= indexOfQFM(array, 7,  5) == 10;
	passed &= indexOfQFM(array, 8,  4) == 10;
	passed &= indexOfQFM(array, 9,  3) == 10;
	passed &= indexOfQFM(array, 10, 2) == 10;
	passed &= indexOfQFM(array, 11, 1) == -1;

	return passed;
}

uint8_t cpyInc_test(void){
	uint8_t passed = 1;

	volatile uint8_t arr0[12] = 
		{0xaa, 0xbb, 0xf1, 0xcc, 
		 0xaa, 0xbb, 0xf1, 0xcc,
		 0xaa, 0xbb, 0xf1, 0xcc};

	volatile uint8_t arr1[12] = 
		{0x00, 0x00, 0x00, 0x00,
		 0x00, 0x00, 0x00, 0x00,
		 0x00, 0x00, 0x00, 0x00};

	int32_t arr0_idx = 0;
	int32_t arr1_idx = 0;

	cpyInc(arr1, arr0, &arr1_idx, &arr0_idx, 2);
	passed &= arr1[0]  == 0xaa;
	passed &= arr1[1]  == 0xbb;
	passed &= arr1[2]  == 0x00;
	passed &= arr1[3]  == 0x00;
	passed &= arr1[4]  == 0x00;
	passed &= arr1[5]  == 0x00;
	passed &= arr1[6]  == 0x00;
	passed &= arr1[7]  == 0x00;
	passed &= arr1[8]  == 0x00;
	passed &= arr1[9]  == 0x00;
	passed &= arr1[10] == 0x00;
	passed &= arr1[11] == 0x00;

	cpyInc(arr1, arr0, &arr1_idx, &arr0_idx, 5);
	passed &= arr1[0]  == 0xaa;
	passed &= arr1[1]  == 0xbb;
	passed &= arr1[2]  == 0xf1;
	passed &= arr1[3]  == 0xcc;
	passed &= arr1[4]  == 0xaa;
	passed &= arr1[5]  == 0xbb;
	passed &= arr1[6]  == 0xf1;
	passed &= arr1[7]  == 0x00;
	passed &= arr1[8]  == 0x00;
	passed &= arr1[9]  == 0x00;
	passed &= arr1[10] == 0x00;
	passed &= arr1[11] == 0x00;

	cpyInc(arr1, arr0, &arr1_idx, &arr0_idx, 5);
	passed &= arr1[0]  == 0xaa;
	passed &= arr1[1]  == 0xbb;
	passed &= arr1[2]  == 0xf1;
	passed &= arr1[3]  == 0xcc;
	passed &= arr1[4]  == 0xaa;
	passed &= arr1[5]  == 0xbb;
	passed &= arr1[6]  == 0xf1;
	passed &= arr1[7]  == 0xcc;
	passed &= arr1[8]  == 0xaa;
	passed &= arr1[9]  == 0xbb;
	passed &= arr1[10] == 0xf1;
	passed &= arr1[11] == 0xcc;

	return passed;
}

uint8_t incrementSMPTE_test(void){
    uint8_t passed = 1;
    
    passed &= assert_SMPTE_time(0, 0, 0, 0, 0);
    incrementSMPTE();
    passed &= assert_SMPTE_time(0, 0, 0, 0, 1);
    
    for (uint8_t i = 0; i < 20; i++){
        incrementSMPTE();
    }
    passed &= assert_SMPTE_time(0, 0, 0, 0, 21);

    //Check carry-over
    clearSMPTE();
	passed &= assert_SMPTE_time(0, 0, 0, 0, 0);
    for (uint8_t i = 0; i < SUBFRAME_RATE - 1; i++){
        incrementSMPTE();
    }
    passed &= assert_SMPTE_time(0, 0, 0, 0, SUBFRAME_RATE-1);
    incrementSMPTE();
    passed &= assert_SMPTE_time(0, 0, 0, 1, 0);
    incrementSMPTE();
    passed &= assert_SMPTE_time(0, 0, 0, 1, 1);
	
	//Check long execution
	clearSMPTE();
	for (uint16_t i = 0; i < FRAME_RATE * SUBFRAME_RATE; i++){
		incrementSMPTE();
	}
	passed &= assert_SMPTE_time(0, 0, 1, 0, 0);

	return passed;   
}

uint8_t isQuarterFrame_test(void){
	uint8_t passed = 1;
	reset_stream();

	for (uint8_t i = 0; i < 40; i++){
		SMPTE_time->sf = i;
		uint8_t result = isQuarterFrame();
		passed &= SMPTE_time->sf % 10 == 0 ? result : !result;
	}

	reset_stream();
	return passed;
}
uint8_t addQuarterFrameMsg_test(void){
	uint8_t passed = 1;
	reset_stream();

	//Real time execution test
	SMPTE_time->sf  = 30;
	SMPTE_time->f   = 0x12;
	SMPTE_time->sec = 0x34;
	SMPTE_time->min = 0x56;
	SMPTE_time->hr  = 0x78;

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x10);
	passed &= (midi_stream[0] == 0xf1); //Status bit
	passed &= (midi_stream[1] == (0x00 | 0x02)); //MSN: qframe_idx, LSN: Lower nibble of frame
	
	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x20);
	passed &= (midi_stream[2] == 0xf1); //Status bit
	passed &= (midi_stream[3] == (0x10 | 0x01)); //MSN: qframe_idx, LSN: Upper nibble of frame

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x30);
	passed &= (midi_stream[4] == 0xf1); //Status bit
	passed &= (midi_stream[5] == (0x20 | 0x04)); //MSN: qframe_idx, LSN: Lower nibble of sec
	
	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x40);
	passed &= (midi_stream[6] == 0xf1); //Status bit
	passed &= (midi_stream[7] == (0x30 | 0x03)); //MSN: qframe_idx, LSN: Upper nibble of sec

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x50);
	passed &= (midi_stream[8] == 0xf1); //Status bit
	passed &= (midi_stream[9] == (0x40 | 0x06)); //MSN: qframe_idx, LSN: Lower nibble of min

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x60);
	passed &= (midi_stream[10] == 0xf1); //Status bit
	passed &= (midi_stream[11] == (0x50 | 0x05)); //MSN: qframe_idx, LSN: Upper nibble of min

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x70);
	passed &= (midi_stream[12] == 0xf1); //Status bit
	passed &= (midi_stream[13] == (0x60 | 0x08)); //MSN: qframe_idx, LSN: Lower nibble of hr

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x00);
	passed &= (midi_stream[14] == 0xf1); //Status bit
	passed &= (midi_stream[15] == (0x70 | 0x07)); //MSN: qframe_idx, LSN: Upper nibble of hr

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x10);
	passed &= (midi_stream[16] == 0xf1); //Status bit
	passed &= (midi_stream[17] == (0x00 | 0x03)); //MSN: qframe_idx, LSN: Lower nibble of frame, increased
	
	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x20);
	passed &= (midi_stream[18] == 0xf1); //Status bit
	passed &= (midi_stream[19] == (0x10 | 0x01)); //MSN: qframe_idx, LSN: Upper nibble of frame

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x30);
	passed &= (midi_stream[20] == 0xf1); //Status bit
	passed &= (midi_stream[21] == (0x20 | 0x04)); //MSN: qframe_idx, LSN: Lower nibble of sec
	
	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x40);
	passed &= (midi_stream[22] == 0xf1); //Status bit
	passed &= (midi_stream[23] == (0x30 | 0x03)); //MSN: qframe_idx, LSN: Upper nibble of sec

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x50);
	passed &= (midi_stream[24] == 0xf1); //Status bit
	passed &= (midi_stream[25] == (0x40 | 0x06)); //MSN: qframe_idx, LSN: Lower nibble of min

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x60);
	passed &= (midi_stream[26] == 0xf1); //Status bit
	passed &= (midi_stream[27] == (0x50 | 0x05)); //MSN: qframe_idx, LSN: Upper nibble of min

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x70);
	passed &= (midi_stream[28] == 0xf1); //Status bit
	passed &= (midi_stream[29] == (0x60 | 0x08)); //MSN: qframe_idx, LSN: Lower nibble of hr

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x00);
	passed &= (midi_stream[30] == 0xf1); //Status bit
	passed &= (midi_stream[31] == (0x70 | 0x07)); //MSN: qframe_idx, LSN: Upper nibble of hr

	//Test independent of quarter frame tick
	SMPTE_time->f   = 0x89;
	SMPTE_time->sec = 0xab;
	SMPTE_time->min = 0xcd;
	SMPTE_time->hr  = 0xef;

	addQuarterFrameMsg();
	passed &= (qframe_idx == 0x10);
	passed &= (midi_stream[32] == 0xf1); //Status bit
	passed &= (midi_stream[33] == (0x00 | 0x09)); //MSN: qframe_idx, LSN: Lower nibble of frame
	
	addQuarterFrameMsg();
	passed &= (qframe_idx == 0x20);
	passed &= (midi_stream[34] == 0xf1); //Status bit
	passed &= (midi_stream[35] == (0x10 | 0x08)); //MSN: qframe_idx, LSN: Upper nibble of frame

	addQuarterFrameMsg();
	quarter_frame_tick();
	passed &= (qframe_idx == 0x30);
	passed &= (midi_stream[36] == 0xf1); //Status bit
	passed &= (midi_stream[37] == (0x20 | 0x0b)); //MSN: qframe_idx, LSN: Lower nibble of sec
	
	addQuarterFrameMsg();
	passed &= (qframe_idx == 0x40);
	passed &= (midi_stream[38] == 0xf1); //Status bit
	passed &= (midi_stream[39] == (0x30 | 0x0a)); //MSN: qframe_idx, LSN: Upper nibble of sec

	addQuarterFrameMsg();;
	passed &= (qframe_idx == 0x50);
	passed &= (midi_stream[40] == 0xf1); //Status bit
	passed &= (midi_stream[41] == (0x40 | 0x0d)); //MSN: qframe_idx, LSN: Lower nibble of min

	addQuarterFrameMsg();
	passed &= (qframe_idx == 0x60);
	passed &= (midi_stream[42] == 0xf1); //Status bit
	passed &= (midi_stream[43] == (0x50 | 0x0c)); //MSN: qframe_idx, LSN: Upper nibble of min

	addQuarterFrameMsg();
	passed &= (qframe_idx == 0x70);
	passed &= (midi_stream[44] == 0xf1); //Status bit
	passed &= (midi_stream[45] == (0x60 | 0x0f)); //MSN: qframe_idx, LSN: Lower nibble of hr

	addQuarterFrameMsg();
	passed &= (qframe_idx == 0x00);
	passed &= (midi_stream[46] == 0xf1); //Status bit
	passed &= (midi_stream[47] == (0x70 | 0x0e)); //MSN: qframe_idx, LSN: Upper nibble of hr

	reset_stream();
	return passed;
}

uint8_t addHeaderChunk_test_helper(void){
	uint8_t passed = 1;

	//Add chunk ID, which is the character string "MThd"
    passed &= midi_stream[0] == 0x4d; //'M'
    passed &= midi_stream[1] == 0x54; //'T'
    passed &= midi_stream[2] == 0x68; //'h'
    passed &= midi_stream[3] == 0x64; //'d'
	//return passed;
    
    //Add the size of the header chunk in bytes, which is 6 by default
    passed &= midi_stream[4] == 0x00;
    passed &= midi_stream[5] == 0x00;
    passed &= midi_stream[6] == 0x00;
    passed &= midi_stream[7] == 0x06;

    //Add the format type, which is default 0 to allow a single track only
	//We will be merging all tracks into one single track for simplicity
    passed &= midi_stream[8] == 0x00;
    passed &= midi_stream[9] == 0x00;
    
    //Add the number of tracks, which we set to 1 by default
    passed &= midi_stream[10] == 0x00;
    passed &= midi_stream[11] == 0x01;
    
    //Add the time division, which will be in frames per second mode
	
	uint8_t neg_frame_rate = ~(FRAME_RATE) + 1; //We lose negative sign in uint8_t array
    passed &= midi_stream[12] == neg_frame_rate; 
    passed &= midi_stream[13] == SUBFRAME_RATE;

	return passed;
}


uint8_t addHeaderChunk_test(void){
	uint8_t passed = 1;
	reset_stream();
	
	addHeaderChunk();
    passed &= addHeaderChunk_test_helper();

	reset_stream();
	return passed;
}

uint8_t addTrackChunk_test_helper(uint8_t idx, uint32_t trk_len){
	uint8_t passed = 1;

    //Add chunk ID, which is the character string "MTrk"
    passed &= midi_stream[idx]   == 0x4d; //'M'
    passed &= midi_stream[idx+1] == 0x54; //'T'
    passed &= midi_stream[idx+2] == 0x72; //'r'
    passed &= midi_stream[idx+3] == 0x6b; //'k'
    
    //Add chunk size
    passed &= get_uint32(midi_stream, idx+4) == trk_len;

	return passed;
}

uint8_t addTrackChunk_test(void){
	uint8_t passed = 1;
	reset_stream();
	
	addTrackChunk(0x12345678);
    passed &= addTrackChunk_test_helper(0, 0x12345678);
	passed &= midi_stream_idx == 8;

	reset_stream();
	return passed;
}

uint8_t finalizeMidiStream_test(void){
	uint8_t passed = 1;
	reset_stream();
	uint16_t test_len = 200;

	//Add data to the MIDI stream
	for (uint16_t i = 0; i < test_len; i++){
		midi_stream[midi_stream_idx++] = i;
	}

	passed &= midi_stream_idx == 200;
	finalizeMidiStream();

	//Verify data was moved over properly
	for (uint16_t i = 0; i < test_len; i++){
		passed &= midi_stream[i + MIDI_STREAM_OVHD] == i;
	}

	//Check proper header was formed
	passed &= addHeaderChunk_test_helper();
	
	//Check proper track was formed
	passed &= addTrackChunk_test_helper(MIDI_STREAM_HEADER_LEN, test_len);

	//Check idx is correct
	passed &= midi_stream_idx == test_len + MIDI_STREAM_OVHD;

	reset_stream();
	return passed;
}


void mergeMidiStreams_test_init(uint8_t new_stream[]){
	//Initializes `midi_stream` with sample data and creates
	//and returns a `new_stream` with sample data

	//First, create sample streams	

	//Add fake MIDI data
	for (int32_t i = 0; i < 128; i++){
		midi_stream[i] = 0xab;
	}
	//Replace every 8th byte with a status byte (as in a QFM)
	for (int32_t i = 0; i < 128; i = i + 8){
		midi_stream[i] = 0xf1;
	} 
	//Add 0 where data byte of QFM should be
	for (int32_t i = 1; i < 128; i = i + 8){
		midi_stream[i] = 0x00;
	}

	midi_stream_idx = 128;
	finalizeMidiStream();

	//Copy the header for the new stream
    //Add chunk ID, which is the character string "MThd"
    new_stream[0] = 0x4d; //'M'
    new_stream[1] = 0x54; //'T'
    new_stream[2] = 0x68; //'h'
    new_stream[3] = 0x64; //'d'
    
    //Add the size of the header chunk in bytes, which is 6 by default
   	new_stream[4] = 0x00;
    new_stream[5] = 0x00;
    new_stream[6] = 0x00;
    new_stream[7] = 0x06;
    
    //Add the format type, which is default 0 to allow a single track only
	//We will be merging all tracks into one single track for simplicity
    new_stream[8] = 0x00;
    new_stream[9] = 0x00;
    
    //Add the number of tracks, which we set to 1 by default
    new_stream[10] = 0x00;
    new_stream[11] = 0x01;
    
    //Add the time division, which will be in frames per second mode
    new_stream[12] = -FRAME_RATE;
    new_stream[13] = SUBFRAME_RATE;

	//Add chunk ID, which is the character string "MTrk"
    new_stream[14] = 0x4d; //'M'
    new_stream[15] = 0x54; //'T'
    new_stream[16] = 0x72; //'r'
    new_stream[17] = 0x6b; //'k'

	//`new_stream` will be length 62
	set_uint32(new_stream, 18, 62);
	
	int32_t start_idx = MIDI_STREAM_OVHD;
	//Add fake MIDI data
	for (int32_t i = 0; i < 64; i++){
		new_stream[i + start_idx] = 0xcd;
	}
	//Replace every 8th byte with a status byte (as in a QFM)
	for (int32_t i = 0; i < 64; i = i + 8){
		new_stream[i + start_idx] = 0xf1;
	} 
	//Add 0 where data byte of QFM should be
	for (int32_t i = 1; i < 64; i = i + 8){
		new_stream[i + start_idx] = 0x00;
	}
}

void mergeMidiStreams_test_init2(uint8_t new_stream[]){
	reset_stream();
	//Add fake MIDI data
	for (int32_t i = 0; i < 128; i++){
		midi_stream[i] = 0xab;
	}
	//Replace every 8th byte with a status byte (as in a QFM)
	for (int32_t i = 0; i < 128; i = i + 8){
		midi_stream[i] = 0xf1;
	} 
	//Add 0 where data byte of QFM should be
	for (int32_t i = 1; i < 128; i = i + 8){
		midi_stream[i] = 0x00;
	}

	midi_stream_idx = 128;
	finalizeMidiStream();

	//`new_stream` will be length 250
	set_uint32(new_stream, 18, 250);
	
	int32_t start_idx = MIDI_STREAM_OVHD;
	//Add fake MIDI data
	for (int32_t i = 0; i < 250; i++){
		new_stream[i + start_idx] = 0xcd;
	}
	//Replace every 8th byte with a status byte (as in a QFM)
	for (int32_t i = 0; i < 250; i = i + 8){
		new_stream[i + start_idx] = 0xf1;
	} 
	//Add 0 where data byte of QFM should be
	for (int32_t i = 1; i < 250; i = i + 8){
		new_stream[i + start_idx] = 0x00;
	}


}

uint8_t mergeMidiStreams_test(void){
	uint8_t passed = 1;
	uint8_t new_stream[300];
	mergeMidiStreams_test_init(new_stream);
	
	passed &= mergeMidiStreams(new_stream);
	passed &= addHeaderChunk_test_helper();

	//Check normal execution
	for (int32_t i = MIDI_STREAM_OVHD; i < MIDI_STREAM_OVHD + 7 * 14; i = i + 14){
		passed &= midi_stream[i]    == 0xf1;
		passed &= midi_stream[i+1]  == 0x00;
		passed &= midi_stream[i+2]  == 0xab;
		passed &= midi_stream[i+3]  == 0xab;
		passed &= midi_stream[i+4]  == 0xab;
		passed &= midi_stream[i+5]  == 0xab;
		passed &= midi_stream[i+6]  == 0xab;
		passed &= midi_stream[i+7]  == 0xab;
		passed &= midi_stream[i+8]  == 0xcd;
		passed &= midi_stream[i+9]  == 0xcd;
		passed &= midi_stream[i+10] == 0xcd;
		passed &= midi_stream[i+11] == 0xcd;
		passed &= midi_stream[i+12] == 0xcd;
		passed &= midi_stream[i+13] == 0xcd;
	}

	//`new_stream` terminates first, so it will come ahead of `midi_stream`
	//in this iteration
	int32_t pivot = MIDI_STREAM_OVHD + 7 * 14;
	passed &= midi_stream[pivot]    == 0xf1;
	passed &= midi_stream[pivot+1]  == 0x00;
	passed &= midi_stream[pivot+2]  == 0xcd;
	passed &= midi_stream[pivot+3]  == 0xcd;
	passed &= midi_stream[pivot+4]  == 0xcd;
	passed &= midi_stream[pivot+5]  == 0xcd;
	passed &= midi_stream[pivot+6]  == 0xab;
	passed &= midi_stream[pivot+7]  == 0xab;
	passed &= midi_stream[pivot+8]  == 0xab;
	passed &= midi_stream[pivot+9]  == 0xab;
	passed &= midi_stream[pivot+10] == 0xab;
	passed &= midi_stream[pivot+11] == 0xab;
	pivot += 12;

	//Check execution after `new_stream` terminates and only
	//`midi_stream` has any remaining data
	for (int32_t i = pivot; i < pivot + 7*8; i = i + 8){
		passed &= midi_stream[i]    == 0xf1;
		passed &= midi_stream[i+1]  == 0x00;
		passed &= midi_stream[i+2]  == 0xab;
		passed &= midi_stream[i+3]  == 0xab;
		passed &= midi_stream[i+4]  == 0xab;
		passed &= midi_stream[i+5]  == 0xab;
		passed &= midi_stream[i+6]  == 0xab;
		passed &= midi_stream[i+7]  == 0xab;
	}

	mergeMidiStreams_test_init2(new_stream);
	passed &= mergeMidiStreams(new_stream);

	//Check normal execution
	for (int32_t i = MIDI_STREAM_OVHD; i < MIDI_STREAM_OVHD + 15 * 14; i = i + 14){
		passed &= midi_stream[i]    == 0xf1;
		passed &= midi_stream[i+1]  == 0x00;
		passed &= midi_stream[i+2]  == 0xab;
		passed &= midi_stream[i+3]  == 0xab;
		passed &= midi_stream[i+4]  == 0xab;
		passed &= midi_stream[i+5]  == 0xab;
		passed &= midi_stream[i+6]  == 0xab;
		passed &= midi_stream[i+7]  == 0xab;
		passed &= midi_stream[i+8]  == 0xcd;
		passed &= midi_stream[i+9]  == 0xcd;
		passed &= midi_stream[i+10] == 0xcd;
		passed &= midi_stream[i+11] == 0xcd;
		passed &= midi_stream[i+12] == 0xcd;
		passed &= midi_stream[i+13] == 0xcd;
	}

	//`midi_stream` terminates first, so it will come ahead of `new_stream`
	//in this iteration
	pivot = MIDI_STREAM_OVHD + 15 * 14;


	passed &= midi_stream[pivot]    == 0xf1;
	passed &= midi_stream[pivot+1]  == 0x00;
	passed &= midi_stream[pivot+2]  == 0xab;
	passed &= midi_stream[pivot+3]  == 0xab;
	passed &= midi_stream[pivot+4]  == 0xab;
	passed &= midi_stream[pivot+5]  == 0xab;
	passed &= midi_stream[pivot+6]  == 0xab;
	passed &= midi_stream[pivot+7]  == 0xab;
	passed &= midi_stream[pivot+8]  == 0xcd;
	passed &= midi_stream[pivot+9]  == 0xcd;
	passed &= midi_stream[pivot+10] == 0xcd;
	passed &= midi_stream[pivot+11] == 0xcd;
	passed &= midi_stream[pivot+12] == 0xcd;
	passed &= midi_stream[pivot+13] == 0xcd;
	pivot += 14;

	//Check execution after `new_stream` terminates and only
	//`midi_stream` has any remaining data
	for (int32_t i = pivot; i < pivot + 7*8; i = i + 8){
		passed &= midi_stream[i]    == 0xf1;
		passed &= midi_stream[i+1]  == 0x00;
		passed &= midi_stream[i+2]  == 0xcd;
		passed &= midi_stream[i+3]  == 0xcd;
		passed &= midi_stream[i+4]  == 0xcd;
		passed &= midi_stream[i+5]  == 0xcd;
		passed &= midi_stream[i+6]  == 0xcd;
		passed &= midi_stream[i+7]  == 0xcd;
	}

	return passed;
}

uint8_t addTimeDelta_test(){
    uint8_t passed = 1;
    reset_stream();

	//Mark unused space in stream with zeros
	for (uint8_t i = 0; i < 8; i++){
		midi_stream[i] = 0;
	}

    addTimeDelta(0x00);
    passed &= midi_stream[0] == (0x00 | 0x00);
	passed &= midi_stream[1] == 0x00;
	passed &= midi_stream_idx == 1;
    reset_stream();
    
    addTimeDelta(0x09);
    passed &= midi_stream[0] == (0x00 | 0x09);
	passed &= midi_stream[1] == 0x00;
	passed &= midi_stream_idx == 1;
    reset_stream();
    
    addTimeDelta(0xff); //1_111,1111
    passed &= midi_stream[0] == (0x80 | 0x01);
    passed &= midi_stream[1] == (0x00 | 0x7f);
	passed &= midi_stream[2] == 0x00;
	passed &= midi_stream_idx == 2;
    reset_stream();
    
    addTimeDelta(0x1234); //00_010,0100_011,0100
    passed &= midi_stream[0] == (0x80 | 0x24);
    passed &= midi_stream[1] == (0x00 | 0x34);
	passed &= midi_stream[2] == 0x00;
	passed &= midi_stream_idx == 2;
    reset_stream();
    
    addTimeDelta(0x364758ff); //0011_011,0010_001,1101_011,0001_111,1111
    passed &= midi_stream[0] == (0x80 | 0x03);
    passed &= midi_stream[1] == (0x80 | 0x32);
    passed &= midi_stream[2] == (0x80 | 0x1d);
    passed &= midi_stream[3] == (0x80 | 0x31);
    passed &= midi_stream[4] == (0x00 | 0x7f);
	passed &= midi_stream[5] == 0x00;
	passed &= midi_stream_idx == 5;
    reset_stream();
    
    return passed;
}

uint8_t changeChannel_test(void){
	uint8_t passed = 1;

	//Create sample stream
	for(uint8_t i = 1; i <= 10; i++){
		addQuarterFrameMsg();
		for(uint8_t j = 0; j < i; j++){
			midiEvent(0x8, 0x4, 0xaa, 0xbb);
		}
	}
	finalizeMidiStream();

	int32_t idx = MIDI_STREAM_OVHD;
	for(uint8_t i = 1; i <= 10; i++){
		idx += 2; //Skip QFM
		for(uint8_t j = 0; j < i; j++){
			passed &= LSB(midi_stream[idx]) == 0x84;
			idx += 3; //Skip to next event
		}
	}

	changeChannel(midi_stream, 0xa, 10*2 + 55*3 + MIDI_STREAM_OVHD);
	idx = MIDI_STREAM_OVHD;
	for(uint8_t i = 1; i <= 10; i++){
		idx += 2; //Skip QFM
		for(uint8_t j = 0; j < i; j++){
			passed &= LSB(midi_stream[idx]) == 0x8a;
			idx += 3; //Skip to next event
		}
	}

	return passed;
}

int main(void){
    uart_init();
	initMidi();
	stdout = stdin = stderr = &uart_str;
    fprintf(stdout, "Initializing midi_lib_test.c\n");
    
    uint8_t test_passed;

	test_passed = LSN_test();
    fprintf(stdout, "LSN_test: %i\n", test_passed);

	test_passed = MSN_test();
    fprintf(stdout, "MSN_test: %i\n", test_passed);
    
    test_passed = get_uint16_test();
    fprintf(stdout, "get_uint16_test: %i\n", test_passed);
    
    test_passed = get_uint32_test();
    fprintf(stdout, "get_uint32_test: %i\n", test_passed);
    
    test_passed = set_uint16_test();
    fprintf(stdout, "set_uint16_test: %i\n", test_passed);
    
    test_passed = set_uint32_test();
    fprintf(stdout, "set_uint32_test: %i\n", test_passed);

	test_passed = indexOfQFM_test();
    fprintf(stdout, "indexOfQFM_test: %i\n", test_passed);

	test_passed = cpyInc_test();
    fprintf(stdout, "cpyInc_test: %i\n", test_passed);

    test_passed = incrementSMPTE_test();
    fprintf(stdout, "incrementSMPTE_test: %i\n", test_passed);

	test_passed = isQuarterFrame_test();
    fprintf(stdout, "isQuarterFrame_test: %i\n", test_passed);

    test_passed = addQuarterFrameMsg_test();
    fprintf(stdout, "addQuarterFrameMsg_test: %i\n", test_passed);

	test_passed = addHeaderChunk_test();
    fprintf(stdout, "addHeaderChunk_test: %i\n", test_passed);

	test_passed = addTrackChunk_test();
    fprintf(stdout, "addTrackChunk_test: %i\n", test_passed);

	test_passed = finalizeMidiStream_test();
    fprintf(stdout, "finalizeMidiStream_test: %i\n", test_passed);

	test_passed = mergeMidiStreams_test();
    fprintf(stdout, "mergeMidiStreams_test: %i\n", test_passed);

	test_passed = addTimeDelta_test();
    fprintf(stdout, "addTimeDelta_test: %i\n", test_passed);

	test_passed = changeChannel_test();
    fprintf(stdout, "changeChannel_test: %i\n", test_passed);

	fprintf(stdout, "\n");
}
